# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
TestRunner负责多个测试用例，目前提供三种方式:

    - 单线程执行 TestRunner
    - 多线程执行 ThreadingTestRunner
    - 多进程执行 MultiProcessTestRunner

整个执行逻辑可以用以下伪代码来理解::

    for test in tests:
        report.begin_test(test)
        result = report.get_testresult_factory().create(test)
        report.end_test(test, result)

"""
from __future__ import absolute_import

import argparse
import collections
import itertools
import multiprocessing
import random
import re
import socket
import sys
import threading
import traceback
import types
import uuid

import six

from six.moves import queue

from testbase import logger
from testbase.loader import TestLoader
from testbase import serialization
from testbase.report import TestReportBase
from testbase.testcase import TestCase, TestCaseRunner
from testbase.testsuite import TestSuiteBase
from testbase.testresult import TestResultCollection, TestResultType
from testbase.resource import TestResourceManager, LocalResourceManagerBackend
from testbase.plan import TestPlan
from testbase.util import ShareDataManager

# pylint: disable=invalid-name
runner_usage = (
    'runtest <test ...> --runner-type <runner-type> [--runner-args "<runner-args>"]'
)


class TestCaseSettings(object):
    """目标测试用例配置"""

    def __init__(
        self,
        names=None,
        excluded_names=None,
        priorities=None,
        status=None,
        owners=None,
        tags=None,
        excluded_tags=None,
        share_data=None,
        global_parameters=None,
        stop=False
    ):
        """构造函数

        :param names: 测试用例名
        :type names: list
        :param excluded_names: 排除测试用例名
        :type excluded_names: list
        :param priorities: 指定用例优先级，如果不指定则默认全部优先级
        :type priorities: list
        :param status: 指定用例状态，如果不指定则默认全部状态
        :type status: list
        :param owners: 指定用户名进行过滤
        :type owners: list
        :param tags: 指定标签选择用例
        :type tags: list
        :param excluded_tags: 指定标签排除用例
        :type tags: list
        :param stop: 遇见执行失败时，是否停止测试，如果True则支持，默认False
        :type stop: bool
        """
        share_data = share_data or {}
        global_parameters = global_parameters or {}
        self.stop = stop
        if names is None:
            self.names = []
        else:
            # self.names = list(names)
            self.names = [
                name
                if isinstance(name, dict)
                else {"name": self._generate_testcase_names(name)}
                for name in names
            ]

        for case in self.names:
            if "parameters" in case:
                for parameter in global_parameters:
                    if parameter not in case["parameters"]:
                        case["parameters"][parameter] = global_parameters[parameter]
            else:
                case["parameters"] = global_parameters

        if owners is None:
            self.owners = []
        else:
            self.owners = list(owners)

        if excluded_names is None:
            excluded_names = []
        else:
            excluded_names = list(excluded_names)
        self.excluded_names = []
        for it in excluded_names:
            if it:
                self.excluded_names.append(it)

        if priorities is None:
            priorities = [
                TestCase.EnumPriority.BVT,
                TestCase.EnumPriority.High,
                TestCase.EnumPriority.Normal,
                TestCase.EnumPriority.Low,
            ]

        if status is None:
            status = [
                TestCase.EnumStatus.Design,
                TestCase.EnumStatus.Implement,
                TestCase.EnumStatus.Ready,
                TestCase.EnumStatus.Review,
                TestCase.EnumStatus.Suspend,
            ]
        self.priorities = priorities
        self.status = status

        self.excluded_cls_names = []
        self.excluded_mod_names = []
        for it in self.excluded_names:
            if self._is_test_class(it):
                self.excluded_cls_names.append(it)
            else:
                self.excluded_mod_names.append(it)

        self.tags = set(tags) if tags else None
        self.excluded_tags = set(excluded_tags) if excluded_tags else None
        self.share_data = share_data

    def _generate_testcase_names(self, testcase):
        if testcase.endswith(".py"):
            testcase = re.sub(r"\\|/", ".", testcase)
            testcase = re.sub(r"\.py$", "", testcase)
        return testcase

    def _is_test_class(self, name):
        """判断路径是否是一个类名"""
        if "/" in name:  # 数据驱动相关的类
            return True
        if "." not in name:  # 不可能直接引用类名
            return False

        try:
            __import__(name)
            return False  # 为模块
        except ImportError:  # pylint: disable=broad-except
            modname, clsname = name.rsplit(".", 1)
            try:
                __import__(modname)
                mod = sys.modules[modname]
            except Exception:  # pylint: disable=broad-except
                return False  # 不存在的模块
            else:
                return hasattr(mod, clsname)

    def filter(self, testcase):
        """测试用例过滤函数

        :param testcase: 测试用例
        :type testcase: TestCase
        """
        name = testcase.test_name
        for it in self.excluded_cls_names:
            if name.startswith(it):
                return "match excluded list: '%s'" % it
        for it in self.excluded_mod_names:
            if name.startswith(it + "."):
                return "match excluded list: '%s'" % it
        if testcase.status not in self.status:
            return "testcase with status '%s' is excluded" % testcase.status
        if testcase.priority not in self.priorities:
            return "testcase with priority '%s' is excluded" % testcase.priority
        if self.owners and testcase.owner not in self.owners:
            return "testcase with owner '%s' is excluded" % testcase.owner
        if self.tags and self.tags.isdisjoint(testcase.tags):
            return "testcase is not tagged with %s" % ("/".join(self.tags))
        if self.excluded_tags and not self.excluded_tags.isdisjoint(testcase.tags):
            return "testcase is tag with %s" % ("/".join(self.excluded_tags))
        return False


class BaseTestRunner(object):
    """测试执行器基类"""

    def __init__(self, report, resmgr_backend=None, execute_type="random"):
        """构造函数

        :param report: 测试报告
        :type report: ITestReport
        """
        self.__report = report
        self.stop = False
        if resmgr_backend is None:
            resmgr_backend = LocalResourceManagerBackend()
        self._resmgr = TestResourceManager(resmgr_backend)
        self._execute_type = execute_type

        self._not_run = []  # 未执行的用例
        self._error_msg = ""    # 错误信息，目前用于未执行的用例

        self._share_data_mgr = ShareDataManager()

    @property
    def report(self):
        """对应的测试报告

        :returns: ITestReport
        """
        return self.__report

    def load_share_data(self, data):
        if not isinstance(data, dict):
            logger.warn(
                "share_data expected dict instance, %s found" % type(data).__name__
            )
            return
        for key, value in data.items():
            self._share_data_mgr.set(key, value)

    def get_share_data(self, key):
        return self._share_data_mgr.get(key)

    def add_share_data(self, key, value, level=0):
        self._share_data_mgr.set(key, value, level)

    def remove_share_data(self, name):
        self._share_data_mgr.remove(name)

    def load(self, target):
        """加载测试用例

        :param target: 指定要执行的测试用例
        :type target: list(TestCase) or list(string) or string or TestCaseSettings
        :returns: 测试用例列表
        """
        if isinstance(target, str):
            target = TestCaseSettings(names=target.split(" "))
        elif (
            isinstance(target, list)
            and target
            and isinstance(target[0], six.string_types)
        ):
            target = TestCaseSettings(names=target)

        if isinstance(target, TestCaseSettings):
            loader = TestLoader(target.filter)
            tests = loader.load(target.names)
            self.__report.log_loaded_tests(loader, tests)
            filtered_tests = loader.get_filtered_tests_with_reason().items()
            for test, reason in filtered_tests:
                self.__report.log_filtered_test(loader, test, reason)
            for testname, error in loader.get_last_errors().items():
                self.__report.log_load_error(loader, testname, error)
            self.__report.info(
                "Loader", "filter %s testcases totally" % len(filtered_tests)
            )
            self.__report.info("Loader", "load %s testcases totally" % len(tests))
        else:
            tests = target
        return tests

    def run(self, target):
        """运行测试

        :param target: 指定要执行的测试
        :type target: list(TestCase) or list(string) or string or TestCaseSettings or TestPlan
        """
        if isinstance(target, TestPlan):
            plan = target
            target = plan.get_tests()
        else:
            self.stop = getattr(target, "stop", False)  # TestCaseSettings stop

            class SimplePlan(TestPlan):
                tests = target
                test_target_args = {}

            plan = SimplePlan()

        if isinstance(target, TestCaseSettings):
            self.load_share_data(target.share_data)

        self.__report.begin_report()
        plan.test_setup(self.__report)
        self.__report.log_test_target(plan.get_test_target())
        self.resource_setup(plan)
        self.run_all_tests(self.load(target))
        if self._not_run:
            for testcase in self._not_run:
                self.__report.log_not_run_test(testcase, TestResultType.FILTERED, self._error_msg)
        self.resource_teardown(plan)
        plan.test_teardown(self.__report)
        self.__report.end_report()
        self.clean_up()
        return self.__report

    def resource_setup(self, plan):
        """资源初始化

        :param plan: 测试计划
        :type plan: TestPlan
        """
        for res_type, resource in itertools.chain(
            [("node", {"host": socket.gethostname(), "id": uuid.getnode()})],
            self._resmgr.iter_managed_resource(),
        ):
            ret_resource = plan.resource_setup(self.__report, res_type, resource)
            if ret_resource:
                resource = ret_resource
            self.__report.log_resource(res_type, resource)

    def resource_teardown(self, plan):
        """资源清理

        :param plan: 测试计划
        :type plan: TestPlan
        """
        for res_type, resource in self._resmgr.iter_managed_resource():
            plan.resource_teardown(self.__report, res_type, resource)
        plan.resource_teardown(
            self.__report, "node", {"host": socket.gethostname(), "id": uuid.getnode()}
        )

    def run_all_tests(self, tests):
        """执行全部的测试用例

        :param tests: 测试用例对象列表
        :type tests: list
        """
        if self._execute_type == "random":
            random.shuffle(tests)
        for test in tests:
            self.run_test(test)

    def run_test(self, test):
        """执行一个测试用例

        :param test: 测试用例
        :type test: TestCase
        :returns: boolean - 测试是否通过
        """
        test.test_resmgr = self._resmgr
        if isinstance(test, TestCase):
            test.share_data_mgr = self._share_data_mgr  # 用于传递共享数据
        elif isinstance(test, TestSuiteBase):
            test.share_data_mgr = self._share_data_mgr
            for it in test:
                it.share_data_mgr = self._share_data_mgr

        runner = getattr(test, "case_runner", TestCaseRunner())
        result = runner.run(test, self.__report.get_testresult_factory())
        if isinstance(result, TestResultCollection):
            self._log_collection_result(result)
        else:
            self.__report.log_test_result(result.testcase, result)
        return result.passed

    def clean_up(self):
        """执行清理动作"""
        pass

    def _log_collection_result(self, result_collection):
        """记录结果集合"""
        for it in result_collection:
            if isinstance(it, TestResultCollection):
                self._log_collection_result(it)
            else:
                self.__report.log_test_result(it.testcase, it)

    @classmethod
    def get_parser(cls):
        """获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        """
        raise NotImplementedError()

    @classmethod
    def parse_args(cls, args_string, report, resmgr_backend, execute_type):
        """通过命令行参数构造对象

        :returns: 测试报告
        :rtype: cls
        """
        raise NotImplementedError()

    def handle_result(self, test, tests_queue, passed, tests_retry_dict):
        if not passed:
            if self.stop:
                if len(tests_queue) <= 0:
                    return
                while len(tests_queue) > 0:
                    testcase = tests_queue.popleft()
                    self._error_msg = "some testcase failed, stop on failure."
                    self._not_run.append(testcase)
                return
            tests_retry_dict.setdefault(test, 0)
            if tests_retry_dict[test] < self._retries:
                tests_retry_dict[test] += 1
                tests_queue.append(test)


class TestRunner(BaseTestRunner):
    """测试执行器"""

    def __init__(self, report, retries=0, resmgr_backend=None, execute_type="random"):
        """构造函数

        :param result: 测试报告
        :type result: ITestReport
        :param retries: 用例失败时重试次数
        :type retries: int
        """
        super(TestRunner, self).__init__(report, resmgr_backend, execute_type)
        self._retries = retries

    def run_all_tests(self, tests):
        """执行全部的测试用例

        :param test: 测试用例对象列表
        :type tests: list
        """
        if self._execute_type == "random":
            random.shuffle(tests)
        tests_queue = collections.deque(tests)
        tests_retry_dict = {}
        while len(tests_queue) > 0:
            test = tests_queue.popleft()
            passed = self.run_test(test)
            self.handle_result(test=test,
                             tests_queue=tests_queue,
                             passed=passed,
                             tests_retry_dict=tests_retry_dict)

    @classmethod
    def get_parser(cls):
        """获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        """
        parser = argparse.ArgumentParser(usage=runner_usage)
        parser.add_argument(
            "--retries", type=int, default=0, help="retry count while test case failed"
        )
        return parser

    @classmethod
    def parse_args(cls, args_string, report, resmgr_backend, execute_type):
        """通过命令行参数构造对象

        :returns: 测试报告
        :rtype: cls
        """
        args = cls.get_parser().parse_args(args_string)
        return cls(report, args.retries, resmgr_backend, execute_type)


class ThreadSafetyReport(TestReportBase):
    """TestReport修饰器，保证线程安全"""

    def __init__(self, report):
        """构造函数

        :param result: 测试报告
        :type result: ITestReport
        """
        super(ThreadSafetyReport, self).__init__()
        self._lock = threading.Lock()
        self._report = report

    def begin_report(self):
        """开始测试执行"""
        with self._lock:
            return self._report.begin_report()

    def end_report(self):
        """结束测试执行

        :param passed: 测试是否通过
        :type passed: boolean
        """
        with self._lock:
            return self._report.end_report()

    def log_test_result(self, testcase, testresult):
        """记录一个测试结果

        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: TestResult
        """
        with self._lock:
            super(ThreadSafetyReport, self).log_test_result(testcase, testresult)
            return self._report.log_test_result(testcase, testresult)

    def log_loaded_tests(self, loader, testcases):
        """记录加载成功的用例

        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcases: 测试用例列表
        :type testcases: list
        """
        with self._lock:
            return self._report.log_loaded_tests(loader, testcases)

    def log_filtered_test(self, loader, testcase, reason):
        """记录一个被过滤的测试用例
        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcase: 测试用例
        :type testcase: TestCase
        :param reason: 过滤原因
        :type reason: str
        """
        with self._lock:
            return self._report.log_filtered_test(loader, testcase, reason)

    def log_load_error(self, loader, name, error):
        """记录一个加载失败的用例或用例集
        :param loader: 用例加载器
        :type loader: TestLoader
        :param name: 名称
        :type name: str
        :param error: 错误信息
        :type error: str
        """
        with self._lock:
            return self._report.log_load_error(loader, name, error)

    def get_testresult_factory(self):
        """获取对应的TestResult工厂

        :returns: ITestResultFactory
        """
        with self._lock:
            return self._report.get_testresult_factory()

    def log_record(self, level, tag, msg, record):
        """增加一个记录

        :param level: 日志级别
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type level: string
        :type tag: string
        :type msg: string
        :type record: dict
        """
        with self._lock:
            return self._report.log_record(level, tag, msg, record)


class ThreadingTestRunner(BaseTestRunner):
    """使用多线程并发执行用例"""

    def __init__(
        self,
        report,
        thread_cnt=0,
        retries=0,
        resmgr_backend=None,
        execute_type="random",
    ):
        """构造函数

        :param report: 测试报告
        :type report: ITestReport
        :param thread_cnt: 线程数
        :type thread_cnt: int
        :param retries: 用例失败时重试次数
        :type retries: int
        """
        self.concurrency = int(thread_cnt) or multiprocessing.cpu_count()
        self._retries = retries
        self._lock = threading.Lock()
        if self.concurrency > 1:
            report = ThreadSafetyReport(report)
        super(ThreadingTestRunner, self).__init__(report, resmgr_backend, execute_type)

    def run_all_tests(self, tests):
        """执行全部的测试用例

        :param test: 测试用例对象列表
        :type tests: list
        """
        if self._execute_type == "random":
            random.shuffle(tests)
        tests_queue = collections.deque(tests)
        tests_retry_dict = {}
        threads = []
        for _ in range(self.concurrency):
            thread = threading.Thread(
                target=self._run_test_from_queue, args=(tests_queue, tests_retry_dict)
            )
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()

    def _run_test_from_queue(self, tests_queue, tests_retry_dict):
        """从队列中不断取用例并执行

        :param tests_queue: 测试用例队列
        :type tests_queue: deque
        :param tests_retry_dict: 测试用例重跑记录
        :type tests_retry_dict: dict
        """
        while len(tests_queue) > 0:
            with self._lock:
                if len(tests_queue) <= 0:
                    break
                test = tests_queue.pop()
            passed = self.run_test(test)
            with self._lock:
                self.handle_result(test=test,
                                 tests_queue=tests_queue,
                                 passed=passed,
                                 tests_retry_dict=tests_retry_dict)


    @classmethod
    def get_parser(cls):
        """获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        """
        parser = argparse.ArgumentParser(usage=runner_usage)
        parser.add_argument(
            "--retries", type=int, default=0, help="retry count while test case failed"
        )
        parser.add_argument(
            "--concurrency", type=int, default=0, help="number of concurrent thread"
        )
        return parser

    @classmethod
    def parse_args(cls, args_string, report, resmgr_backend, execute_type):
        """通过命令行参数构造对象

        :returns: 测试报告
        :rtype: cls
        """
        args = cls.get_parser().parse_args(args_string)
        return cls(report, args.concurrency, args.retries, resmgr_backend, execute_type)


class EnumProcessMsgType(object):
    """多进程间通信用的消息类型"""

    Worker_Quit = 0
    Worker_RunTest = 1
    Worker_Idle = 2
    Worker_Error = 3

    Result_GetAttr = 4
    Result_AttrValue = 5
    Result_AttrError = 6
    Result_Func = 7
    Result_CallFunc = 8
    Result_Return = 9
    Result_Raise = 10

    Report_LogTestResult = 12
    Report_LogRecord = 13


class TestResultFunctionProxy(object):
    """测试结果函数代理"""

    def __init__(self, from_worker, obj_id, func_name):
        """构造函数

        :param from_worker: 所属的工作者
        :type from_worker: TestWorker
        :param obj_id: 对象ID
        :type obj_id: int
        :param func_name: 函数名
        :type func_name: string
        """
        self._worker = from_worker
        self._objid = obj_id
        self._func_name = func_name

    def __call__(self, *args, **kwargs):
        self._worker.send_message(
            (
                EnumProcessMsgType.Result_CallFunc,
                self._objid,
                self._func_name,
                args,
                kwargs,
            )
        )
        msg = self._worker.recv_message(5)
        msg_type = msg[0]
        if msg_type == EnumProcessMsgType.Result_Return:
            return msg[1]
        elif msg_type == EnumProcessMsgType.Result_Raise:
            raise RuntimeError(str(msg[1]))
        else:
            raise RuntimeError("unexpect message: %s" % msg)


class TestResultProxy(object):
    """测试结果代理"""

    def __init__(self, from_worker, obj_id, passed, testcase):
        """构造函数

        :param from_worker: 来源的工作者
        :type from_worker: TestWorker
        :param obj_id: 对象ID
        :type obj_id: int
        :param passed: 测试是否通过
        :type passed: boolean
        :param testcase: 对应测试用例
        :type testcase: TestCase
        """
        self.__worker = from_worker
        self.__passed = passed
        self.__objid = obj_id
        self.__testcase = testcase

    def __getattr__(self, name):
        if name.startswith("_TestResultProxy__"):
            return super(TestResultProxy, self).__getattr__(name)
        if name == "passed":  # shortcut
            return self.__passed
        if name == "testcase":
            return self.__testcase
        self.__worker.send_message(
            (EnumProcessMsgType.Result_GetAttr, self.__objid, name)
        )
        msg = self.__worker.recv_message(5)
        msg_type = msg[0]
        if msg_type == EnumProcessMsgType.Result_AttrValue:
            return msg[1]
        elif msg_type == EnumProcessMsgType.Result_Func:
            return TestResultFunctionProxy(self.__worker, self.__objid, name)
        elif EnumProcessMsgType.Result_AttrError:
            raise AttributeError(msg[1])
        else:
            raise RuntimeError("unexpect message: %s" % msg)

    def __setattr__(self, name, value):
        if name.startswith("_TestResultProxy__"):
            super(TestResultProxy, self).__setattr__(name, value)
        else:
            raise RuntimeError("read only")


class TestReportProxy(TestReportBase):
    """测试报告代理"""

    def __init__(self, worker_id, ctrl_msg_queue, result_factory, result_manager):
        """构造函数

        :param worker_id: 工作者ID
        :type worker_id: string
        :param ctrl_msg_queue: 控制进程的消息队列
        :type ctrl_msg_queue: multiprocessing.Queue
        :param result_factory: 测试结果工厂
        :type result_factory: ITestResultFactory
        :param result_manager: 测试结果残根管理器
        :type result_manager：TestResultStubManager
        """
        super(TestReportProxy, self).__init__()
        self._worker_id = worker_id
        self._ctrl_msg_queue = ctrl_msg_queue
        self._result_factory = result_factory
        self._result_manager = result_manager

    def begin_report(self):
        """开始测试执行"""
        raise RuntimeError("should not call this")

    def end_report(self):
        """结束测试执行

        :param passed: 测试是否通过
        :type passed: boolean
        """
        raise RuntimeError("should not call this")

    def log_test_result(self, testcase, testresult):
        """记录一个测试结果

        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: TestResult
        """
        super(TestReportProxy, self).log_test_result(testcase, testresult)
        objid = self._result_manager.add_result(testresult)
        self._ctrl_msg_queue.put(
            (
                EnumProcessMsgType.Report_LogTestResult,
                self._worker_id,
                serialization.dumps(testcase),
                objid,
                testresult.passed,
            )
        )

    def log_record(self, level, tag, msg, record):
        """增加一个记录

        :param level: 日志级别
        :param tag: 日志标签
        :param msg: 日志消息
        :param record: 日志记录信息
        :type level: string
        :type tag: string
        :type msg: string
        :type record: dict
        """
        self._ctrl_msg_queue.put(
            (EnumProcessMsgType.Report_LogRecord, (level, tag, msg, record))
        )

    def get_testresult_factory(self):
        """获取对应的TestResult工厂

        :returns: ITestResultFactory
        """
        return self._result_factory


class _EmptyClass(object):
    pass


class TestResultStubManager(object):
    """测试结果桩管理器"""

    def __init__(self, rsp_queue):
        """构造函数

        :param rsp_queue: 对工作者请求结果的答复消息队列
        :type rsp_queue:  multiprocessing.Queue
        """
        self._rsp_queue = rsp_queue
        self._result_dict = {}

    def add_result(self, result):
        """增加一个测试结果"""
        self._result_dict[id(result)] = result
        return id(result)

    def get_result_attr(self, objid, attrname):
        """获取一个测试结果的属性值

        :param objid: 对象ID
        :type objid: int
        :param attrname: 属性名
        :type attrname: string
        """
        result = self._result_dict[objid]
        try:
            attr = getattr(result, attrname)
            if not isinstance(attr, types.MethodType):
                rsp = EnumProcessMsgType.Result_AttrValue, attr
            else:
                rsp = (EnumProcessMsgType.Result_Func,)
            self._rsp_queue.put(rsp)
        except Exception:  # pylint: disable=broad-except
            self._rsp_queue.put(
                (
                    EnumProcessMsgType.Result_AttrError,
                    "'%s' object has no attribute '%s'"
                    % (type(result).__name__, attrname),
                )
            )

    def call_result_func(self, objid, funcname, args, kwargs):
        """调用一个测试结果的函数

        :param objid: 对象ID
        :type objid: int
        :param funcname: 函数名
        :type funcname: string
        :param args: 参数
        :type args: tuple
        :param kwargs: 参数
        :type kwargs: dict
        """
        result = self._result_dict[objid]
        try:
            rsp = EnumProcessMsgType.Result_Return, getattr(result, funcname)(
                *args, **kwargs
            )
        except Exception:  # pylint: disable=broad-except
            rsp = EnumProcessMsgType.Result_Raise, traceback.format_exc()
        self._rsp_queue.put(rsp)


def _log_collection_result(testreport, result_collection):
    """记录结果集合"""
    for it in result_collection:
        if isinstance(it, TestResultCollection):
            _log_collection_result(testreport, it)
        else:
            testreport.log_test_result(it.testcase, it)


def _run_test_thread(
    worker_id, ctrl_msg_queue, testcase, testreport, resmgr, share_data_mgr=None
):
    """执行测试用例的线程

    :param worker_id: 工作者ID
    :type worker_id: string
    :param ctrl_msg_queue: 控制进程的消息队列
    :type ctrl_msg_queue: multiprocessing.Queue
    :param testcase: 测试用例
    :type testcase: TestCase
    :param testreport: 测试报告
    :type testreport: ITestReport
    :param resmgr: 资源管理器
    :type resmgr: TestResourceManager
    """
    try:
        case_runner = getattr(testcase, "case_runner", TestCaseRunner())
        testcase.test_resmgr = resmgr
        testcase.share_data_mgr = share_data_mgr
        result = case_runner.run(testcase, testreport.get_testresult_factory())
        if isinstance(result, TestResultCollection):
            _log_collection_result(testreport, result)
        else:
            testreport.log_test_result(result.testcase, result)

        ctrl_msg_queue.put(
            (
                EnumProcessMsgType.Worker_Idle,
                worker_id,
                serialization.dumps(testcase),
                result.passed,
            )
        )
    except Exception:  # pylint: disable=broad-except
        ctrl_msg_queue.put(
            (EnumProcessMsgType.Worker_Error, worker_id, traceback.format_exc())
        )


def _worker_process(
    worker_id,
    ctrl_msg_queue,
    msg_queue,
    rsp_queue,
    result_factory_type,
    result_factory_data,
    resmgr,
    share_data_mgr=None,
):
    """执行测试的子进程过程

    :param worker_id: 工作者ID，全局唯一
    :type worker_id: string
    :param ctrl_msg_queue: 控制进程通信的消息队列
    :type ctrl_msg_queue: multiprocessing.Queue
    :param msg_queue: 本进程的消息队列
    :type msg_queue: multiprocessing.Queue
    :param rsp_queue: 对本进程请求结果的答复消息队列
    :type rsp_queue:  multiprocessing.Queue
    :param result_factory_type: 测试结果工厂类
    :type result_factory_type: type
    :param result_factory_data: 测试结果工厂序列化后数据
    :type result_factory_data: object
    :param resmgr: 资源管理器
    :type resmgr: TestResourceManager
    """
    try:
        result_factory = _EmptyClass()
        result_factory.__class__ = result_factory_type
        result_factory.loads(result_factory_data)
        result_manager = TestResultStubManager(rsp_queue)
        report = TestReportProxy(
            worker_id, ctrl_msg_queue, result_factory, result_manager
        )
        while True:
            msg = msg_queue.get()
            msg_type = msg[0]
            msg_data = msg[1:]

            if msg_type == EnumProcessMsgType.Worker_Quit:
                break

            elif msg_type == EnumProcessMsgType.Worker_RunTest:
                testcase = serialization.loads(msg_data[0])
                t = threading.Thread(
                    target=_run_test_thread,
                    args=(
                        worker_id,
                        ctrl_msg_queue,
                        testcase,
                        report,
                        resmgr,
                        share_data_mgr,
                    ),
                )
                t.daemon = True
                t.start()

            elif msg_type == EnumProcessMsgType.Result_GetAttr:
                objid, name = msg_data
                result_manager.get_result_attr(objid, name)

            elif msg_type == EnumProcessMsgType.Result_CallFunc:
                objid, func, args, kwargs = msg_data
                result_manager.call_result_func(objid, func, args, kwargs)
    except Exception:  # pylint: disable=broad-except
        ctrl_msg_queue.put(
            (EnumProcessMsgType.Worker_Error, worker_id, traceback.format_exc())
        )


class TestWorker(object):
    """多进程执行用例时，执行测试的子进程"""

    def __init__(
        self, worker_id, ctrl_msg_queue, result_factory, resmgr, share_data_mgr=None
    ):
        """构造函数

        :param worker_id: 工作者ID，全局唯一
        :type worker_id: string
        :param ctrl_msg_queue: 控制进程的消息队列
        :type msg_queue: multiprocessing.Queue
        :param result_factory: 测试结果工厂
        :type result_factory: ITestResultFactory
        :param resmgr: 资源管理器
        :type resmgr: TestResourceManager
        """
        self._worker_id = worker_id
        self._result_factory = result_factory
        self._ctrl_msg_queue = ctrl_msg_queue
        self._resmgr = resmgr
        self._share_data_mgr = share_data_mgr
        self._reset()

    def _reset(self):
        """重置内部状态"""
        self._rsp_queue = multiprocessing.Queue()
        self._msg_queue = multiprocessing.Queue()
        self._process = multiprocessing.Process(
            target=_worker_process,
            args=(
                self._worker_id,
                self._ctrl_msg_queue,
                self._msg_queue,
                self._rsp_queue,
                type(self._result_factory),
                self._result_factory.dumps(),
                self._resmgr,
                self._share_data_mgr,
            ),
        )
        self._monitor = threading.Thread(target=self._process_monitor)
        self._monitor.daemon = True
        self._testcase = None
        self._stopping = False

    def _process_monitor(self):
        """监控线程"""
        self._process.join()
        if not self._stopping:
            self._ctrl_msg_queue.put(
                (
                    EnumProcessMsgType.Worker_Error,
                    self._worker_id,
                    "process exit unexpectedly",
                )
            )

    def start(self):
        """开始执行"""
        self._process.start()
        self._monitor.start()

    def stop(self):
        """结束执行"""
        self._stopping = True
        self.send_message((EnumProcessMsgType.Worker_Quit,))
        self._process.join(5)
        if self._process.is_alive():
            self._process.terminate()

    def restart(self):
        """重新开始执行"""
        self._reset()
        self.start()

    def run_testcase(self, testcase):
        """分配一个测试用例

        :param testcase: 要执行的测试用例
        :type testcase: TestCase
        """
        self.send_message(
            (EnumProcessMsgType.Worker_RunTest, serialization.dumps(testcase))
        )
        self._testcase = testcase

    def current_testcase(self):
        """当前正在执行的测试用例

        :returns: TestCase
        """
        return self._testcase

    def send_message(self, msg):
        """发送消息到工作者

        :param msg: 消息
        :type msg: tuple
        """
        self._msg_queue.put(msg)

    def recv_message(self, timeout=None):
        """接收工作者的答复消息"""
        if timeout is None:
            return self._rsp_queue.get()
        else:
            try:
                return self._rsp_queue.get(timeout=timeout)
            except queue.Empty:
                raise RuntimeError("waiting response message from worker timeout")


class MultiProcessTestRunner(BaseTestRunner):
    """使用多进程并发执行用例

    多进程并发时，有两个特殊的问题需要处理：

    1、测试执行工作进程需要通知TestReport测试用例的执行情况等，
    解决方案是：
    为每个工作进程提供一个TestReportProxy，TestReportProxy通过消息机制通知
    真正的TestReport

    2、TestReport需要访问在工作进程的TestResult对象，
    解决方案是：
    每个工作进程有一个TestResultStubManager，提供给TestReport的是一个TestResultProxy
    对象，TestResultProxy通过消息机制和TestResultStubManager通信，来获取真正的TestResult
    的信息

    """

    def __init__(
        self,
        report,
        process_cnt=0,
        retries=0,
        resmgr_backend=None,
        execute_type="random",
    ):
        """构造函数

        :param report: 测试报告
        :type report: ITestReport
        :param process_cnt: 进程数
        :type process_cnt: int
        :param retries: 失败重跑次数上限
        :type retries: int
        """
        self.concurrency = int(process_cnt) or multiprocessing.cpu_count()
        self._retries = retries
        self._workers_dict = {}
        super(MultiProcessTestRunner, self).__init__(
            report, resmgr_backend, execute_type
        )

        self._share_data_mgr = ShareDataManager(
            lock=multiprocessing.Lock(), data=multiprocessing.Manager().dict()
        )

    def run_all_tests(self, tests):
        """执行全部的测试用例

        :param test: 测试用例对象列表
        :type tests: list
        """
        if len(tests) == 0:
            return
        if len(tests) < self.concurrency:
            self.concurrency = len(tests)

        if self._execute_type == "random":
            random.shuffle(tests)
        tests_queue = collections.deque(tests)
        tests_retry_dict = {}
        msg_queue = multiprocessing.Queue()
        result_factory = self.report.get_testresult_factory()
        for i in range(self.concurrency):
            worker = TestWorker(
                i, msg_queue, result_factory, self._resmgr, self._share_data_mgr
            )
            worker.start()
            worker.run_testcase(tests_queue.popleft())
            self._workers_dict[i] = worker

        idle_workers = []
        error_counts = {}
        while True:
            msg = msg_queue.get()
            msg_type = msg[0]

            if msg_type == EnumProcessMsgType.Report_LogTestResult:
                worker = self._workers_dict[msg[1]]
                testcase = serialization.loads(msg[2])
                objid, passed = msg[3], msg[4]
                self.report.log_test_result(
                    testcase, TestResultProxy(worker, objid, passed, testcase)
                )

            elif msg_type == EnumProcessMsgType.Report_LogRecord:
                self.report.log_record(msg[1], msg[2], msg[3], msg[4])

            elif msg_type == EnumProcessMsgType.Worker_Idle:
                worker = self._workers_dict[msg[1]]
                test = serialization.loads(msg[2])
                passed = msg[3]
                if not passed:
                    if self.stop:   # 如果当前有用例执行失败，配置了stop_on_failure时，停止测试
                        for index in list(self._workers_dict.keys())[::-1]:
                            _worker = self._workers_dict[index]
                            current_case = _worker.current_testcase()
                            if worker != _worker:
                                self._not_run.append(current_case)
                                self._error_msg = "some testcase failed, stop on failure."  # 其它进程的用例也进入not_run list
                        if len(tests_queue) <= 0:
                            return
                        while len(tests_queue) > 0:
                            testcase = tests_queue.popleft()
                            self._error_msg = "some testcase failed, stop on failure."
                            self._not_run.append(testcase)
                        return
                    tests_retry_dict.setdefault(test.test_name, 0)
                    if tests_retry_dict[test.test_name] < self._retries:
                        tests_retry_dict[test.test_name] += 1
                        tests_queue.append(test)

                if len(tests_queue) > 0:
                    worker.run_testcase(tests_queue.popleft())
                else:
                    idle_workers.append(worker)
                    if len(idle_workers) == len(self._workers_dict):
                        break

            elif msg_type == EnumProcessMsgType.Worker_Error:
                if msg[1] not in self._workers_dict:
                    continue
                worker = self._workers_dict[msg[1]]
                err_msg = msg[2]
                self.report.error(
                    "RUNNER",
                    "runner process %s error occurred: %s" % (msg[1], err_msg),
                    record=dict(err_msg=err_msg),
                )
                error_count = error_counts.get(msg[1], 0)
                if error_count >= 4:
                    self.report.error(
                        "RUNNER",
                        "worker with id=%s failed to run for up 3 times" % msg[1],
                    )
                    worker.stop()
                    del self._workers_dict[msg[1]]
                    if len(idle_workers) == len(self._workers_dict):
                        break
                else:
                    error_counts[msg[1]] = error_count + 1
                    if worker not in idle_workers:
                        tests_queue.append(worker.current_testcase())
                        worker.restart()
                        worker.run_testcase(tests_queue.popleft())

    def clean_up(self):
        BaseTestRunner.clean_up(self)
        for it in self._workers_dict.values():
            it.stop()

    @classmethod
    def get_parser(cls):
        """获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        """
        parser = argparse.ArgumentParser(usage=runner_usage)
        parser.add_argument(
            "--retries", type=int, default=0, help="retry count while test case failed"
        )
        parser.add_argument(
            "--concurrency", type=int, default=0, help="number of concurrent thread"
        )
        return parser

    @classmethod
    def parse_args(cls, args_string, report, resmgr_backend, execute_type):
        """通过命令行参数构造对象

        :returns: 测试报告
        :rtype: cls
        """
        args = cls.get_parser().parse_args(args_string)
        return cls(report, args.concurrency, args.retries, resmgr_backend, execute_type)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""log模块
"""

import logging
import sys
import traceback
from testbase import context
from testbase.util import ensure_binary_stream, smart_binary

_stream, _encoding = ensure_binary_stream(sys.stdout)

class _Formatter(logging.Formatter):
    def format(self, record):
        s = super(_Formatter, self).format(record)
        return smart_binary(s, encoding=_encoding)


_stream_handler = logging.StreamHandler(_stream)
_stream_handler.terminator = b"\n"
_stream_handler.setFormatter(_Formatter())


class TestResultBridge(logging.Handler):
    """中转log信息到TestResult"""

    def emit(self, log_record):
        """Log Handle 必须实现此函数"""
        testresult = context.current_testresult()
        if testresult is None:
            _stream_handler.emit(log_record)
            return
        record = {}
        if log_record.exc_info:
            record["traceback"] = "".join(
                traceback.format_tb(log_record.exc_info[2])
            ) + "%s: %s" % (log_record.exc_info[0].__name__, log_record.exc_info[1])
        testresult.log_record(log_record.levelno, log_record.msg, record)


_LOGGER_NAME = "QTA_LOGGER"
_logger = logging.getLogger(_LOGGER_NAME)
_logger.setLevel(logging.DEBUG)
_logger.addHandler(TestResultBridge())


def critical(msg, *args, **kwargs):
    _logger.error(msg, *args, **kwargs)


fatal = critical


def error(msg, *args, **kwargs):
    """Log a message with severity 'ERROR' on the root logger."""
    _logger.error(msg, *args, **kwargs)


def exception(msg, *args):
    """Log a message with severity 'ERROR' on the root logger,with exception information."""
    _logger.exception(msg, *args)


def warning(msg, *args, **kwargs):
    """Log a message with severity 'WARNING' on the root logger."""
    _logger.warning(msg, *args, **kwargs)


warn = warning


def info(msg, *args, **kwargs):
    """Log a message with severity 'INFO' on the root logger."""
    _logger.info(msg, *args, **kwargs)


def debug(msg, *args, **kwargs):
    """Log a message with severity 'DEBUG' on the root logger."""
    _logger.debug(msg, *args, **kwargs)


def log(level, msg, *args, **kwargs):
    """Log 'msg % args' with the integer severity 'level' on the root logger."""
    _logger.log(level, msg, *args, **kwargs)


def addHandler(hdlr):  # pylint: disable=invalid-name
    """Add the specified handler to this logger."""
    _logger.addHandler(hdlr)


def removeHandler(hdlr):  # pylint: disable=invalid-name
    """Remove the specified handler from this logger."""
    _logger.removeHandler(hdlr)

def set_formatter(fmt):
    """Set the specified formatter to this logger.
    """
    class __formatter(_Formatter):
        def __init__(self, fmt):
            super(_Formatter, self).__init__(fmt)

    _stream_handler.setFormatter(__formatter(fmt))

def set_level(level):
    """Set the specified log level to this logger.
    """
    _logger.setLevel(level)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""测试项目接口
"""

from __future__ import unicode_literals

import os
import datetime
import zipfile
import getpass
import shutil
import xml.dom.minidom as dom

from testbase.util import codecs_open

INITPY_CONTENT = """# -*- coding: utf-8 -*-
'''%(Doc)s
'''
#%(Date)s QTAF自动生成
"""

SETTINGS_CONTENT_STANDALONE = """# -*- coding: utf-8 -*-
'''测试项目配置文件
'''
#%(Date)s QTAF自动生成

PROJECT_NAME = "%(ProjectName)s"
PROJECT_MODE = "standalone"
"""

SETTINGS_CONTENT_STANDARD = """# -*- coding: utf-8 -*-
'''测试项目配置文件
'''
#%(Date)s QTAF自动生成

import os

PROJECT_NAME = "%(ProjectName)s"
PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
PROJECT_MODE = "standard"
INSTALLED_APPS = []
"""

TESTCASE_CONTENT = """# -*- coding: utf-8 -*-
'''示例测试用例
'''
#%(Date)s QTAF自动生成

from %(ProjectName)slib.testcase import %(ProjectNameCapUp)sTestCase


class HelloTest(%(ProjectNameCapUp)sTestCase):
    '''示例测试用例
    '''
    owner = "%(UserName)s"
    timeout = 5
    priority = %(ProjectNameCapUp)sTestCase.EnumPriority.High
    status = %(ProjectNameCapUp)sTestCase.EnumStatus.Design

    def run_test(self):
        self.log_info("hello testcase")
        self.assert_equal(True, True)

if __name__ == '__main__':
    HelloTest().debug_run()
"""

TESTLIB_CONTENT = """# -*- coding: utf-8 -*-
'''示例测试用例
'''
#%(Date)s QTAF自动生成

from testbase import testcase


class %(ProjectNameCapUp)sTestCase(testcase.TestCase):
    '''%(ProjectName)s测试用例基类
    '''
    pass
"""

MANAGE_CONTENT = """# -*- coding: utf-8 -*-
'''
项目管理和辅助工具
'''
#%(Date)s QTAF自动生成

import sys
import os

PROJ_ROOT = os.path.dirname(os.path.abspath(__file__))
if PROJ_ROOT not in sys.path:
    sys.path.insert(0, PROJ_ROOT)
EXLIB_DIR = os.path.join(PROJ_ROOT, 'exlib')
if os.path.isdir(EXLIB_DIR):
    for filename in os.listdir(EXLIB_DIR):
        if filename.endswith('.egg'):
            lib_path = os.path.join(EXLIB_DIR, filename)
            if os.path.isfile(lib_path) and lib_path not in sys.path:
                sys.path.insert(0, lib_path)

from testbase.management import ManagementTools

if __name__ == '__main__':
    ManagementTools().run()
"""

ECLIPSE_PROJ_CONTENT = """<?xml version="1.0" encoding="UTF-8"?>
<projectDescription>
    <name>%(ProjectName)stestproj</name>
    <comment></comment>
    <projects>
    </projects>
    <buildSpec>
        <buildCommand>
            <name>org.python.pydev.PyDevBuilder</name>
            <arguments>
            </arguments>
        </buildCommand>
    </buildSpec>
    <natures>
        <nature>org.python.pydev.pythonNature</nature>
    </natures>
</projectDescription>
"""

PYDEV_CONF_CONTENT_STANDALONE = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<?eclipse-pydev version="1.0"?><pydev_project>
<pydev_property name="org.python.pydev.PYTHON_PROJECT_INTERPRETER">Default</pydev_property>
<pydev_property name="org.python.pydev.PYTHON_PROJECT_VERSION">python 2.7</pydev_property>
<pydev_pathproperty name="org.python.pydev.PROJECT_SOURCE_PATH">
<path>/${PROJECT_DIR_NAME}</path>
<path>/${PROJECT_DIR_NAME}/exlib/%(EggName)s</path>
</pydev_pathproperty>
</pydev_project>
"""

PYDEV_CONF_CONTENT_STANDARD = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<?eclipse-pydev version="1.0"?><pydev_project>
<pydev_property name="org.python.pydev.PYTHON_PROJECT_INTERPRETER">Default</pydev_property>
<pydev_property name="org.python.pydev.PYTHON_PROJECT_VERSION">python 2.7</pydev_property>
<pydev_pathproperty name="org.python.pydev.PROJECT_SOURCE_PATH">
<path>/${PROJECT_DIR_NAME}</path>
</pydev_pathproperty>
</pydev_project>
"""


class EnumProjectMode(object):
    """测试项目运行模式"""

    Standard = "standard"  # Python标准模式，QTAF和扩展库安装到Python Libs中
    Standalone = "standalone"  # 独立模式，QTAF和扩展库安装到测试项目的exlib目录中


def _create_initpy(dir_path, doc):
    """创建一个__init__.py"""
    with codecs_open(os.path.join(dir_path, "__init__.py"), "wb") as fd:
        fd.write(
            (
                INITPY_CONTENT
                % {"Doc": doc, "Date": datetime.date.today().strftime("%Y/%m/%d")}
            ).encode("utf8")
        )


def _create_settingspy(proj_path, proj_name, mode):
    """创建settings.py文件"""
    if mode == EnumProjectMode.Standalone:
        content = SETTINGS_CONTENT_STANDALONE
    else:
        content = SETTINGS_CONTENT_STANDARD
    with codecs_open(os.path.join(proj_path, "settings.py"), "wb") as fd:
        fd.write(
            (
                content
                % {
                    "Date": datetime.date.today().strftime("%Y/%m/%d"),
                    "ProjectName": proj_name,
                }
            ).encode("utf8")
        )


def _create_managepy(proj_path):
    """创建manage.py文件"""
    with codecs_open(os.path.join(proj_path, "manage.py"), "wb") as fd:
        fd.write(
            (
                MANAGE_CONTENT % {"Date": datetime.date.today().strftime("%Y/%m/%d")}
            ).encode("utf8")
        )


def _create_sample_test(dir_path, proj_name):
    """创建示例测试用例"""
    with codecs_open(os.path.join(dir_path, "hello.py"), "wb") as fd:
        fd.write(
            (
                TESTCASE_CONTENT
                % {
                    "Date": datetime.date.today().strftime("%Y/%m/%d"),
                    "ProjectName": proj_name,
                    "ProjectNameCapUp": proj_name[0].upper() + proj_name[1:],
                    "UserName": getpass.getuser(),
                }
            ).encode("utf8")
        )


def _create_sample_lib(dir_path, proj_name):
    """创建示例lib"""
    with codecs_open(os.path.join(dir_path, "testcase.py"), "wb") as fd:
        fd.write(
            (
                TESTLIB_CONTENT
                % {
                    "Date": datetime.date.today().strftime("%Y/%m/%d"),
                    "ProjectName": proj_name,
                    "ProjectNameCapUp": proj_name[0].upper() + proj_name[1:],
                }
            ).encode("utf8")
        )


def _copy_qtaf_egg(egg_path, dir_path):
    """拷贝QTAF egg包"""
    shutil.copy(egg_path, dir_path)
    with zipfile.ZipFile(egg_path) as zfile:
        try:
            zfile.getinfo("doc/qtaf.chm")  # chm可能不存在
        except KeyError:
            return
        else:
            with codecs_open(os.path.join(dir_path, "qtaf.chm"), "wb") as fd:
                fd.write(zfile.read("doc/qtaf.chm"))


def _create_eclipse_projfile(proj_path, proj_name):
    """创建eclispe项目文件"""
    with codecs_open(os.path.join(proj_path, ".project"), "wb") as fd:
        fd.write((ECLIPSE_PROJ_CONTENT % {"ProjectName": proj_name}).encode("utf8"))


def _create_pydev_conffile(proj_path, mode):
    """创建pydev配置文件"""
    with codecs_open(os.path.join(proj_path, ".pydevproject"), "wb") as fd:
        if mode == EnumProjectMode.Standalone:
            qtaf_egg_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), ".."
            )
            qtaf_egg_path = os.path.abspath(qtaf_egg_path)
            fd.write(
                (
                    PYDEV_CONF_CONTENT_STANDALONE
                    % {"EggName": os.path.basename(qtaf_egg_path)}
                ).encode("utf8")
            )
        else:
            fd.write(PYDEV_CONF_CONTENT_STANDARD.encode("utf8"))


def _update_pydev_conffile(proj_path, egg_name):
    """更新pydev配置文件中的Egg名称"""
    with codecs_open(
        os.path.join(proj_path, ".pydevproject"), "r", encoding="utf-8"
    ) as fd:
        doc = dom.parse(fd)
        nodes = doc.getElementsByTagName("pydev_pathproperty")
        if len(nodes) == 0:
            propsnode = doc.createElement("pydev_pathproperty")
            propsnode.setAttribute("name", "org.python.pydev.PROJECT_SOURCE_PATH")
        else:
            propsnode = nodes[0]
            for pathnode in propsnode.getElementsByTagName("path"):
                rc = []
                for node in pathnode.childNodes:
                    if node.nodeType == node.TEXT_NODE:
                        rc.append(node.data)
                path = "".join(rc)
                if path.endswith(".egg") and "qtaf-" in path:
                    propsnode.removeChild(pathnode)
                    break
        pathnode = doc.createElement("path")
        pathnode.appendChild(
            doc.createTextNode("/${PROJECT_DIR_NAME}/exlib/%s" % egg_name)
        )
        propsnode.appendChild(pathnode)

    with codecs_open(os.path.join(proj_path, ".pydevproject"), "wb", "utf8") as fd:
        fd.write(doc.toxml(encoding="UTF-8"))


def create_project(dest_path, proj_name, mode):
    """创建项目"""
    if not os.path.isdir(dest_path):
        os.makedirs(dest_path)

    lib_dir = os.path.join(dest_path, "%slib" % proj_name)
    if not os.path.isdir(lib_dir):
        os.mkdir(lib_dir)
        _create_initpy(lib_dir, "测试lib")
        _create_sample_lib(lib_dir, proj_name)

    case_dir = os.path.join(dest_path, "%stest" % proj_name)
    if not os.path.isdir(case_dir):
        os.mkdir(case_dir)
        _create_initpy(case_dir, "测试用例")
        _create_sample_test(case_dir, proj_name)

    res_dir = os.path.join(dest_path, "resources")
    if not os.path.isdir(res_dir):
        os.mkdir(res_dir)

    if mode == EnumProjectMode.Standalone:
        exlib_dir = os.path.join(dest_path, "exlib")
        if not os.path.isdir(exlib_dir):
            os.mkdir(exlib_dir)
            qtaf_egg_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), ".."
            )
            qtaf_egg_path = os.path.abspath(qtaf_egg_path)
            _copy_qtaf_egg(qtaf_egg_path, exlib_dir)

    _create_settingspy(dest_path, proj_name, mode)
    _create_managepy(dest_path)
    _create_eclipse_projfile(dest_path, proj_name)
    _create_pydev_conffile(dest_path, mode)


def update_project_qtaf(proj_path, qtaf_egg_path):
    """升级项目的QTAF"""
    exlib_dir = os.path.join(proj_path, "exlib")
    for name in os.listdir(exlib_dir):
        if name.startswith("qtaf-") and name.endswith(".egg"):
            os.remove(os.path.join(exlib_dir, name))
    _copy_qtaf_egg(qtaf_egg_path, exlib_dir)
    _update_pydev_conffile(proj_path, os.path.basename(qtaf_egg_path))


class Project(object):
    """一个项目（废弃接口，存在是为了兼容）"""

    def __init__(self, root):
        self._root = root

    @property
    def path(self):
        return self._root


def current_project():
    """获取当前项目（废弃接口，存在是为了兼容）"""
    from testbase.conf import settings

    return Project(settings.PROJECT_ROOT)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""重试机制实现
"""

import time


class _RetryItem(object):
    """ """

    def __init__(self, iteration, timestamps=None):
        self.__iteration = iteration
        self.__ts = timestamps

    @property
    def iteration(self):
        return self.__iteration

    @property
    def ts(self):
        return self.__ts

    def __str__(self):
        return "<%s iter=%s, ts=%s>" % (
            self.__class__.__name__,
            self.__iteration,
            self.__ts,
        )


class _RetryWithTimeout(object):
    """retry mechanism with timeout"""

    def __init__(self, interval=5, timeout=60, raise_error=True):
        self.interval = interval
        self.timeout = timeout
        self.raise_error = raise_error
        self.__start_time = None
        self.__count = 0

    def __iter__(self):
        return self

    def next(self):
        if self.__start_time is None:
            self.__count += 1
            self.__start_time = time.time()
            return _RetryItem(self.__count, self.__start_time)
        else:
            time.sleep(self.interval)
            ts = time.time()
            if ts - self.__start_time < self.timeout:
                self.__count += 1
                return _RetryItem(self.__count, ts)
            else:
                if self.raise_error:
                    raise RetryLimitExcceeded(
                        "Procedure retried %s times in %ss"
                        % (self.__count, self.timeout)
                    )
                else:
                    raise StopIteration

    __next__ = next


class _RetryWithCount(object):
    """retry mechanism with count"""

    def __init__(self, limit=3, interval=0, raise_error=True):
        self.limit = limit
        self.raise_error = raise_error
        self.interval = interval
        self.__count = 0

    def __iter__(self):
        return self

    def next(self):
        self.__count += 1
        if self.__count <= self.limit:
            if self.__count != 1 and self.interval:
                time.sleep(self.interval)
            return _RetryItem(self.__count, time.time())
        if self.raise_error:
            raise RetryLimitExcceeded(
                "Procedure retried for %s times with interval=%ss"
                % (self.limit, self.interval)
            )
        else:
            raise StopIteration

    __next__ = next


class RetryLimitExcceeded(Exception):
    """maxmium retry limit excceeded error"""

    pass


class Retry(object):
    """retry mechanism with timeout or limit"""

    def __init__(self, timeout=10, limit=None, interval=0.5, raise_error=True):
        if limit:
            self._retry = _RetryWithCount(
                limit=limit, interval=interval, raise_error=raise_error
            )
        else:
            self._retry = _RetryWithTimeout(
                timeout=timeout, interval=interval, raise_error=raise_error
            )

    def __iter__(self):
        return self._retry.__iter__()

    def call(self, callee, *args, **kwargs):
        if not hasattr(callee, "__call__"):
            raise ValueError("callee must be a function instance")
        for _ in self:
            r = callee(*args, **kwargs)
            if r:
                return r



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
测试结果模块

参考logging模块的设计，使用示例如下::

    result = TestResult()
    result.add_handler(StreamResultHandler())
    result.begin_test()
    result.start_step('a step')
    result.info('test')
    result.end_test()

其中result.info接口可以传record扩展信息，比如::

    result.error('', '异常发生', traceback=traceback.format_exc())

同logger一样，TestResult对象保证对所有的ITestResultHandler的调用都是线程安全的，可以通过实现
ITestResultHandler来实现一个新的Handler，详细请参考ITestResultHandler接口

"""

import json
import locale
import os
import sys
import time
import threading
import traceback
import xml.dom.minidom as dom
import xml.parsers.expat as xmlexpat
import xml.sax.saxutils as saxutils

import six

from testbase import context
from testbase.util import (
    TimeoutLock,
    smart_text,
    get_thread_traceback,
    get_method_defined_class,
    to_pretty_xml,
    smart_binary,
    ensure_binary_stream,
    codecs_open,
    get_time_str,
    translate_bad_char,
    path_exists,
)


os_encoding = locale.getdefaultlocale()[1]


class EnumLogLevel(object):
    """日志级别"""

    DEBUG = 10
    INFO = 20
    Environment = 21  # 测试环境相关信息， device/devices表示使用的设备、machine表示执行的机器
    ENVIRONMENT = Environment
    RESOURCE = 22

    WARNING = 30
    ERROR = 40
    ASSERT = 41  # 断言失败，actual/expect/code_location
    CRITICAL = 60
    APPCRASH = 61  # 测试目标Crash
    TESTTIMEOUT = 62  # 测试执行超时
    RESNOTREADY = 69  # 当前资源不能满足测试执行的要求


class TestResultType(object):
    """扩展用例状态"""

    FILTERED = "ignored"


levelname = {}
for name in EnumLogLevel.__dict__:
    value = EnumLogLevel.__dict__[name]
    if isinstance(value, int):
        levelname[value] = name

RESULT_TYPES = ((TestResultType.FILTERED, "被忽略"),)


def _convert_timelength(sec):
    h = int(sec / 3600)
    sec -= h * 3600
    m = int(sec / 60)
    sec -= m * 60
    return (h, m, sec)


def smart_text_by_lines(s):
    """将任意字符串转换为UTF-8编码"""
    lines = []
    for line in s.split("\n"):
        lines.append(smart_text(line))
    return "\n".join(lines)


class TestResultBase(object):
    """测试结果基类

    此类的职责如下：
    1、提供测试结果基本接口
    2、保证线程安全
    2、测试是否通过之逻辑判断
    """

    def __init__(self):
        """构造函数"""
        self.__lock = threading.RLock()
        if sys.version_info[0] >= 3:
            # Support timeout lock on python3
            self.__lock = TimeoutLock(30)
        self.__steps_passed = [True]  # 预设置一个，以防用例中没调用startStep
        self.__curr_step = 0
        self.__accept_result = False
        self.__testcase = None
        self.__begin_time = None
        self.__end_time = None
        self.__error_level = 0
        self.__failed_info = ""
        self.__failed_priority = 0
        self._custom_result = None
        self._custom_reason = None
        self.__failed_stages = []

    @property
    def testcase(self):
        """对应的测试用例
        :returns: TestCase
        """
        return self.__testcase

    @property
    def passed(self):
        """测试是否通过

        :returns: True or False
        """
        return all(self.__steps_passed)

    @property
    def failed_reason(self):
        """用例测试不通过的错误原因

        :returns: str
        """
        if self.__error_level:
            return levelname.get(self.__error_level, "unknown")
        else:
            return ""

    @property
    def failed_info(self):
        """测试用例失败时的执行状况

        :returns: str
        """
        return self.__failed_info

    @property
    def begin_time(self):
        """测试用例开始时间

        :returns: float
        """
        return self.__begin_time

    @property
    def end_time(self):
        """测试用例结束时间

        :returns: float
        """
        return self.__end_time

    @property
    def failed_stages(self):
        """测试用例失败的阶段

        :returns: str
        """
        return self.__failed_stages

    def is_testsuite(self):
        """是否是测试用例套

        :returns: bool
        """
        from testbase.testsuite import TestSuite

        return isinstance(self.testcase, TestSuite)

    def add_failed_stage(self, failed_stage):
        """增加测试用例失败的阶段

        :param value: 阶段名
        :type value: str
        """
        if isinstance(failed_stage, str):
            self.failed_stages.append(failed_stage)

    def begin_test(self, testcase):
        """开始执行测试用例

        :param testcase: 测试用例
        :type testcase: TestCase
        """
        with self.__lock:
            if self.__accept_result:
                raise RuntimeError("此时不可调用begin_test")
            self.__accept_result = True
            self.__begin_time = time.time()
            self.__testcase = testcase
            self.handle_test_begin(testcase)

    def end_test(self):
        """结束执行测试用例"""
        with self.__lock:
            if not self.__accept_result:
                raise RuntimeError("此时不可调用end_test")
            self.handle_step_end(self.__steps_passed[self.__curr_step])
            self.__end_time = time.time()
            self.handle_test_end(self.passed)  # 防止没有一个步骤
            self.__accept_result = False

    def begin_step(self, msg):
        """开始一个测试步骤

        :param msg: 测试步骤名称
        :type msg: string
        """
        with self.__lock:
            if not self.__accept_result:
                raise RuntimeError("此时不可调用begin_step")
            if len(self.__steps_passed) != 1:
                self.handle_step_end(self.__steps_passed[self.__curr_step])
            self.__steps_passed.append(True)
            self.__curr_step += 1
            self.handle_step_begin(msg)

    def log_record(self, level, msg, record=None, attachments=None):
        """处理一个日志记录

        :param level: 日志级别，参考EnumLogLevel
        :type level: string
        :param msg: 日志消息
        :type msg: string
        :param record: 日志记录
        :type record: dict
        :param attachments: 附件
        :type attachments: dict
        """
        if record is None:
            record = {}
        if attachments is None:
            attachments = {}
        if not isinstance(msg, six.string_types):
            raise ValueError("msg='%r'必须是string类型" % msg)
        msg = smart_text(msg)
        if level >= EnumLogLevel.ERROR:
            if self.__testcase.current_stage:
                self.add_failed_stage(self.__testcase.current_stage)
            self.__steps_passed[self.__curr_step] = False
            if level > self.__error_level:
                self.__error_level = level
            extra_record, extra_attachments = self._get_extra_fail_record_safe()
            record.update(extra_record)
            attachments.update(extra_attachments)

            if self.__failed_priority <= 3 and level == EnumLogLevel.APPCRASH:
                self.__failed_info, self.__failed_priority = "Crash", 3

            if self.__failed_priority <= 2 and level == EnumLogLevel.TESTTIMEOUT:
                self.__failed_info, self.__failed_priority = "用例执行超时", 2

            if self.__failed_priority <= 1 and level == EnumLogLevel.ASSERT:
                self.__failed_info, self.__failed_priority = msg, 1

            if self.__failed_priority <= 1 and "traceback" in record:
                if not self.__failed_info:  # 优先记录第一个异常，第一个异常往往比较大可能是问题的原因
                    self.__failed_info, self.__failed_priority = (
                        record["traceback"].split("\n")[-2],
                        1,
                    )

        with self.__lock:
            if not self.__accept_result:
                return
            self.handle_log_record(level, msg, record, attachments)

    def _get_extra_fail_record_safe(self, timeout=300):
        """使用线程调用测试用例的get_extra_fail_record"""

        def _run(outputs, errors):
            try:
                outputs.append(context.current_testcase().get_extra_fail_record())
            except Exception:  # pylint: disable=broad-except
                errors.append(traceback.format_exc())

        errors = []
        outputs = []
        t = threading.Thread(target=_run, args=(outputs, errors))
        t.daemon = True
        t.start()
        t.join(timeout)
        extra_record, extra_attachments = {}, {}
        with self.__lock:
            if t.is_alive():
                stack = get_thread_traceback(t)
                self.handle_log_record(
                    EnumLogLevel.ERROR,
                    "测试失败时获取其他额外错误信息超过了指定时间：%ds" % timeout,
                    {"traceback": stack},
                    {},
                )
            else:
                if errors:
                    self.handle_log_record(
                        EnumLogLevel.ERROR,
                        "测试失败时获取其他额外错误信息失败",
                        {"traceback": errors[0]},
                        {},
                    )
                else:
                    record_info = outputs[0]
                    if isinstance(record_info, (tuple, list)) and len(record_info) == 2:
                        extra_record, extra_attachments = record_info
                    else:
                        cls = get_method_defined_class(
                            self.testcase.get_extra_fail_record
                        )
                        if cls.__module__ == "__main__":
                            class_path = cls.__name__
                        else:
                            class_path = "%s.%s" % (cls.__module__, cls.__name__)
                        raise RuntimeError(
                            "%s.get_extra_fail_record must return a 2 elements tuple"
                            % class_path
                        )
        return extra_record, extra_attachments

    def debug(self, msg, record=None, attachments=None):
        """处理一个DEBUG日志"""
        self.log_record(EnumLogLevel.DEBUG, msg, record, attachments)

    def info(self, msg, record=None, attachments=None):
        """处理一个INFO日志"""
        self.log_record(EnumLogLevel.INFO, msg, record, attachments)

    def warning(self, msg, record=None, attachments=None):
        """处理一个WARNING日志"""
        self.log_record(EnumLogLevel.WARNING, msg, record, attachments)

    def error(self, msg, record=None, attachments=None):
        """处理一个ERROR日志"""
        self.log_record(EnumLogLevel.ERROR, msg, record, attachments)

    def exception(self, msg, record=None, attachments=None):
        """处理一个DEBUG日志"""
        if record is None:
            record = {}
        record["traceback"] = traceback.format_exc()
        self.log_record(EnumLogLevel.CRITICAL, msg, record, attachments)

    def handle_test_begin(self, testcase):
        """处理一个测试用例执行的开始

        :param testcase: 测试用例
        :type testcase: TestCase
        """
        pass

    def handle_test_end(self, passed):
        """处理一个测试用例执行的结束

        :param passed: 测试用例是否通过
        :type passed: boolean
        """
        pass

    def handle_step_begin(self, msg):
        """处理一个测试步骤的开始

        :param msg: 测试步骤名称
        :type msg: string
        """
        pass

    def handle_step_end(self, passed):
        """处理一个测试步骤的结束

        :param passed: 测试步骤是否通过
        :type passed: boolean
        """
        pass

    def handle_log_record(self, level, msg, record, attachments):
        """处理一个日志记录

        :param level: 日志级别，参考EnumLogLevel
        :type level: string
        :param msg: 日志消息
        :type msg: string
        :param record: 日志记录
        :type record: dict
        :param attachments: 附件
        :type attachments: dict
        """
        pass

    def customize_result(self, result, reason=None):
        self._custom_result = result
        if reason:
            self._custom_reason = reason


class EmptyResult(TestResultBase):
    """不输出"""

    pass


class StreamResult(TestResultBase):
    """测试用例stream输出"""

    _seperator1 = "-" * 40 + "\n"
    _seperator2 = "=" * 60 + "\n"

    def __init__(self, stream=sys.stdout):
        """构造函数

        :param stream: 流对象
        :type stream: file
        """
        super(StreamResult, self).__init__()
        self._stream, encoding = ensure_binary_stream(stream)
        self._write = lambda x: self._stream.write(smart_binary(x, encoding=encoding))
        self._step_results = []

    def handle_test_begin(self, testcase):
        """处理一个测试用例执行的开始

        :param testcase: 测试用例
        :type testcase: TestCase
        """
        self._write(self._seperator2)
        owner = getattr(testcase, "owner", None)
        priority = getattr(testcase, "priority", None)
        timeout = getattr(testcase, "timeout", None)
        begin_msg = "测试%s:%s 所有者:%s 优先级:%s 超时:%s分钟\n" % (
            "套" if self.is_testsuite() else "用例",
            testcase.test_name,
            owner,
            priority,
            timeout,
        )

        self._write(begin_msg)
        self._write(self._seperator2)

    def handle_test_end(self, passed):
        """处理一个测试用例执行的结束

        :param passed: 测试用例是否通过
        :type passed: boolean
        """
        self._write(self._seperator2)
        name = "用例"
        if self.is_testsuite():
            name = "套"

        self._write(
            "测试%s开始时间: %s\n"
            % (
                name,
                time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.begin_time)),
            )
        )
        self._write(
            "测试%s结束时间: %s\n"
            % (name, time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.end_time)))
        )
        items = list(_convert_timelength(self.end_time - self.begin_time))
        items.insert(0, name)
        self._write("测试%s执行时间: %02d:%02d:%02.2f\n" % tuple(items))

        rsttxts = {True: "通过", False: "失败"}
        test_result = rsttxts[passed]
        if self._custom_result:
            test_result = dict(RESULT_TYPES)[self._custom_result]
        steptxt = ""
        for i, ipassed in enumerate(self._step_results):
            steptxt += " %s:%s" % (i + 1, rsttxts[ipassed])
        self._write("测试%s步骤结果: %s\n" % (name, steptxt))
        self._write("测试%s最终结果: %s\n" % (name, test_result))
        self._write(self._seperator2)

    def handle_step_begin(self, msg):
        """处理一个测试步骤的开始

        :param msg: 测试步骤名称
        :type msg: string
        """
        if not isinstance(msg, six.string_types):
            raise ValueError("msg='%r'必须是string类型" % msg)

        self._write(self._seperator1)
        self._write("步骤%s: %s\n" % (len(self._step_results) + 1, msg))

    def handle_step_end(self, passed):
        """处理一个测试步骤的结束

        :param passed: 测试步骤是否通过
        :type passed: boolean
        """
        self._step_results.append(passed)

    def handle_log_record(self, level, msg, record, attachments):
        """处理一个日志记录

        :param level: 日志级别，参考EnumLogLevel
        :type level: string
        :param msg: 日志消息
        :type msg: string
        :param record: 日志记录
        :type record: dict
        :param attachments: 附件
        :type attachments: dict
        """
        self._write("%s: %s\n" % (levelname[level], msg))

        if level == EnumLogLevel.ASSERT:
            if "actual" in record:
                actual = record["actual"]
                self._write("   实际值：%s%s\n" % (actual.__class__, actual))
            if "expect" in record:
                expect = record["expect"]
                self._write("   期望值：%s%s\n" % (expect.__class__, expect))
            if "code_location" in record:
                self._write(
                    smart_text(
                        '  File "%s", line %s, in %s\n' % record["code_location"]
                    )
                )

        if "traceback" in record:
            self._write(smart_text_by_lines("%s\n" % record["traceback"]))

        for name in attachments:
            file_path = smart_text(attachments[name])
            if path_exists(file_path):
                file_path = os.path.abspath(file_path)
            self._write("   %s:%s\n" % (smart_text(name), file_path))


class XmlResult(TestResultBase):
    """xml格式的测试用例结果"""

    def __init__(self, testcase):
        """构造函数

        :param file_path: XML文件路径
        :type file_path: string
        """
        super(XmlResult, self).__init__()
        self._xmldoc = dom.Document()
        translated_name = translate_bad_char(testcase.test_name)
        max_name_len = 200
        if len(translated_name) > max_name_len:
            translated_name = translated_name[:max_name_len]
        self._file_path = "%s_%s.xml" % (translated_name, get_time_str())

    @property
    def file_path(self):
        """xml文件路径

        :returns: str
        """
        return self._file_path

    def handle_test_begin(self, testcase):
        """处理一个测试用例执行的开始

        :param testcase: 测试用例
        :type testcase: TestCase
        """
        self._xmldoc.appendChild(
            self._xmldoc.createProcessingInstruction(
                "xml-stylesheet", 'type="text/xsl" href="TestResult.xsl"'
            )
        )
        owner = getattr(testcase, "owner", None)
        priority = getattr(testcase, "priority", None)
        timeout = getattr(testcase, "timeout", None)
        self._testnode = self._xmldoc.createElement("TEST")
        self._testnode.setAttribute(
            "name", smart_text(saxutils.escape(testcase.test_name))
        )
        self._testnode.setAttribute("owner", smart_text(saxutils.escape(str(owner))))
        self._testnode.setAttribute("priority", str(priority))
        self._testnode.setAttribute("timeout", str(timeout))
        self._testnode.setAttribute(
            "begintime",
            time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.begin_time)),
        )
        extra_properties = testcase.get_test_extra_properties()
        for k, v in extra_properties.items():
            self._testnode.setAttribute(k, str(v))
        if hasattr(testcase, "tags"):
            tag_str = "|".join(testcase.tags)
            if tag_str:
                tag_str = "|%s|" % tag_str
            self._testnode.setAttribute("tags", smart_text(saxutils.escape(tag_str)))
        self._xmldoc.appendChild(self._testnode)

        self.begin_step("测试用例初始步骤")

    def handle_test_end(self, passed):
        """处理一个测试用例执行的结束

        :param passed: 测试用例是否通过
        :type passed: boolean
        """
        test_result = str(passed)
        if self._custom_result:
            test_result = str(self._custom_result)
        self._testnode.setAttribute("result", test_result)
        if self._custom_reason:
            self._testnode.setAttribute("reason", self._custom_reason)
        self._testnode.setAttribute(
            "endtime", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.end_time))
        )
        self._testnode.setAttribute(
            "duration",
            "%02d:%02d:%02.2f\n" % _convert_timelength(self.end_time - self.begin_time),
        )
        if self.failed_stages:
            self._testnode.setAttribute("failed_stages", "|".join(self.failed_stages))
        if self._file_path:
            with codecs_open(smart_text(self._file_path), "wb") as fd:
                fd.write(to_pretty_xml(self._xmldoc))

    def handle_step_begin(self, msg):
        """处理一个测试步骤的开始

        :param msg: 测试步骤名称
        :type msg: string
        """
        if not isinstance(msg, six.string_types):
            raise ValueError("msg='%r'必须是string类型" % msg)
        self._stepnode = self._xmldoc.createElement("STEP")
        self._stepnode.setAttribute("title", smart_text(msg))
        self._stepnode.setAttribute(
            "time", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
        )
        self._testnode.appendChild(self._stepnode)

    def handle_step_end(self, passed):
        """处理一个测试步骤的结束

        :param passed: 测试步骤是否通过
        :type passed: boolean
        """
        self._stepnode.setAttribute("result", str(passed))

    def handle_log_record(self, level, msg, record, attachments):
        """处理一个日志记录

        :param level: 日志级别，参考EnumLogLevel
        :type level: string
        :param msg: 日志消息
        :type msg: string
        :param record: 日志记录
        :type record: dict
        :param attachments: 附件
        :type attachments: dict
        """
        if not isinstance(msg, six.string_types):
            msg = str(msg)

        # 由于目前的报告系统仅支持部分级别的标签，所以这里先做转换
        if level >= EnumLogLevel.ERROR:
            tagname = levelname[EnumLogLevel.ERROR]
        elif level == EnumLogLevel.Environment or level == EnumLogLevel.RESOURCE:
            tagname = levelname[EnumLogLevel.INFO]
        else:
            tagname = levelname[level]

        infonode = self._xmldoc.createElement(tagname)
        textnode = self._xmldoc.createTextNode(smart_text(msg))
        infonode.appendChild(textnode)
        self._stepnode.appendChild(infonode)

        if level == EnumLogLevel.ASSERT:
            if "actual" in record:
                node = self._xmldoc.createElement("ACTUAL")
                try:
                    actual = record["actual"]
                    if isinstance(actual, six.string_types):
                        actual = smart_text(actual)
                        dom.parseString("<a>%s</a>" % actual)
                    acttxt = "%s%s" % (actual.__class__, actual)
                except xmlexpat.ExpatError:
                    acttxt = "%s%s" % (actual.__class__, repr(actual))
                except UnicodeEncodeError:
                    acttxt = "%s%s" % (actual.__class__, repr(actual))

                node.appendChild(self._xmldoc.createTextNode(acttxt))
                infonode.appendChild(node)

            if "expect" in record:
                node = self._xmldoc.createElement("EXPECT")
                try:
                    expect = record["expect"]
                    if isinstance(expect, six.string_types):
                        expect = smart_text(expect)
                        dom.parseString("<a>%s</a>" % expect)
                    exptxt = "%s%s" % (expect.__class__, expect)
                except xmlexpat.ExpatError:
                    exptxt = "%s%s" % (expect.__class__, repr(expect))
                except UnicodeEncodeError:
                    exptxt = "%s%s" % (expect.__class__, repr(expect))
                node.appendChild(self._xmldoc.createTextNode(exptxt))
                infonode.appendChild(node)

        if "traceback" in record:
            excnode = self._xmldoc.createElement("EXCEPT")
            excnode.appendChild(
                self._xmldoc.createTextNode(smart_text(record["traceback"]))
            )
            infonode.appendChild(excnode)

        for name in attachments:
            file_path = attachments[name]
            attnode = self._xmldoc.createElement("ATTACHMENT")
            attnode.setAttribute("filepath", smart_text(file_path))
            attnode.appendChild(self._xmldoc.createTextNode(smart_text(name)))
            infonode.appendChild(attnode)

    def toxml(self):
        """返回xml文本

        :returns string - xml文本
        """
        return to_pretty_xml(self._xmldoc)


class JSONResult(TestResultBase):
    """JSON格式的结果"""

    def __init__(self, testcase):
        super(JSONResult, self).__init__()
        self._steps = []
        self._data = {
            "testcase": testcase.test_name,
            "steps": self._steps,
            "failed_info": "",
        }
        self._translated_name = translate_bad_char(testcase.test_name)

    def get_data(self):
        return self._data

    def get_file(self):
        file_name = "%s_%s.json" % (self._translated_name, get_time_str())
        if not path_exists(file_name):
            content = json.dumps(self._data)
            with codecs_open(file_name, mode="w", encoding="utf-8") as fd:
                fd.write(content)
        return file_name

    def handle_test_begin(self, testcase):
        """处理一个测试用例执行的开始

        :param testcase: 测试用例
        :type testcase: TestCase
        """
        self.begin_step("测试用例初始化步骤")

    def handle_test_end(self, passed):
        """处理一个测试用例执行的结束

        :param passed: 测试用例是否通过
        :type passed: boolean
        """
        rsttxts = {True: "通过", False: "失败"}
        test_result = rsttxts[passed]
        if self._custom_result:
            test_result = dict(RESULT_TYPES)[self._custom_result]
        self._data["result_type"] = test_result
        self._data["succeed"] = passed
        self._data["start_time"] = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(self.begin_time)
        )
        self._data["end_time"] = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(self.end_time)
        )
        self._data["failed_info"] = self.failed_info

    def handle_step_begin(self, msg):
        """处理一个测试步骤的开始

        :param msg: 测试步骤名称
        :type msg: string
        """
        self._steps.append(
            {
                "name": msg,
                "start_time": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
                "logs": [],
            }
        )

    def handle_step_end(self, passed):
        """处理一个测试步骤的结束

        :param passed: 测试步骤是否通过
        :type passed: boolean
        """
        curr_step = self._steps[-1]
        curr_step["succeed"] = passed
        curr_step["end_time"] = (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),)

    def handle_log_record(self, level, msg, record, attachments):
        """处理一个日志记录

        :param level: 日志级别，参考EnumLogLevel
        :type level: string
        :param msg: 日志消息
        :type msg: string
        :param record: 日志记录
        :type record: dict
        :param attachments: 附件
        :type attachments: dict
        """
        curr_step = self._steps[-1]
        curr_step["logs"].append(
            {
                "timestamp": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
                "level": level,
                "message": msg,
                "record": record,
                "attachments": attachments,
            }
        )


class HtmlResult(JSONResult):
    """html test result"""

    def get_file(self):
        file_name = "%s_%s.js" % (self._translated_name, get_time_str())
        if not path_exists(file_name):
            var_name = os.path.basename(file_name)
            var_name = os.path.splitext(file_name)[0].replace(".", "_")
            content = "var %s = %s" % (var_name, json.dumps(self._data))
            content = smart_binary(content)
            with codecs_open(file_name, mode="wb") as fd:
                fd.write(content)
        return file_name


class TestResultCollection(list):
    """测试结果集合"""

    def __init__(self, results, passed):
        """构造函数

        :param results: 测试结果列表
        :type results: list
        :param passed: 测试是否通过
        :type passed: boolean
        """
        super(TestResultCollection, self).__init__(results)
        self.__passed = passed

    @property
    def passed(self):
        """测试是否通过

        :returns: boolean
        """
        return self.__passed



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
QTAF配置文件
"""
# -----------------------------------
# 调试模式开关
# -----------------------------------
DEBUG = False

# -----------------------------------
# 全局数据驱动配置
# -----------------------------------
DATA_DRIVE = False
DATA_SOURCE = "test/data/server.py"

# -----------------------------------
# 项目配置
# -----------------------------------
PROJECT_NAME = "qtaf"
PROJECT_MODE = "standalone"  # choices: standard/standalone
PROJECT_ROOT = None  # os.path.dirname(__file__)
INSTALLED_APPS = []


# -----------------------------------
# Assert
# -----------------------------------
QTAF_REWRITE_ASSERT = True



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""qtaf related types
"""

import traceback

import pkg_resources

from testbase import runner, report, resource, logger

RUNNER_ENTRY_POINT = "qtaf.runner"
REPORT_ENTRY_POINT = "qtaf.report"
RESMGR_BACKEND_ENTRY_POINT = "qtaf.resmgr_backend"

runner_types = {}
report_types = {}
resmgr_backend_types = {}


def __init_runner_types():
    global runner_types  # pylint: disable=invalid-name
    if runner_types:
        return
    runner_types["basic"] = runner.TestRunner
    runner_types["multithread"] = runner.ThreadingTestRunner
    runner_types["multiprocess"] = runner.MultiProcessTestRunner
    for ep in pkg_resources.iter_entry_points(RUNNER_ENTRY_POINT):
        if ep.name not in runner_types:
            try:
                runner_types[ep.name] = ep.load()
            except Exception:  # pylint: disable=broad-except
                stack = traceback.format_exc()
                logger.warn(
                    "load TestRunner type for %s failed:\n%s" % (ep.name, stack)
                )


def __init_report_types():
    global report_types  # pylint: disable=invalid-name
    if report_types:
        return
    report_types.update(
        {
            "empty": report.EmptyTestReport,
            "stream": report.StreamTestReport,
            "xml": report.XMLTestReport,
            "json": report.JSONTestReport,
            "html": report.HtmlTestReport,
        }
    )

    # Register other `ITestReport` implementations from entry points
    for ep in pkg_resources.iter_entry_points(REPORT_ENTRY_POINT):
        if ep.name not in report_types:
            try:
                report_types[ep.name] = ep.load()
            except Exception:  # pylint: disable=broad-except
                stack = traceback.format_exc()
                logger.warn(
                    "load ITestReport entry point for %s failed:\n%s" % (ep.name, stack)
                )


def __init_resmgr_backend_types():
    global resmgr_backend_types  # pylint: disable=invalid-name
    if resmgr_backend_types:
        return
    resmgr_backend_types["local"] = resource.LocalResourceManagerBackend
    for ep in pkg_resources.iter_entry_points(RESMGR_BACKEND_ENTRY_POINT):
        if ep.name not in resmgr_backend_types:
            resmgr_backend_types[ep.name] = ep.load()


__init_runner_types()
del __init_runner_types

__init_report_types()
del __init_report_types

__init_resmgr_backend_types()
del __init_resmgr_backend_types



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

import codecs
import os
from setuptools import setup, find_packages

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def generate_version():
    version = "1.0.0"
    if os.path.isfile(os.path.join(BASE_DIR, "version.txt")):
        with codecs.open("version.txt", "r", encoding="utf-8") as fd:
            content = fd.read().strip()
            if content:
                version = content
    with codecs.open(
        os.path.join(BASE_DIR, "testbase", "version.py"), "w", encoding="utf-8"
    ) as fd:
        fd.write('version = "%s"\n' % version)
    return str(version)


def parse_requirements():
    reqs = []
    if os.path.isfile(os.path.join(BASE_DIR, "requirements.txt")):
        with codecs.open(
            os.path.join(BASE_DIR, "requirements.txt"), "r", encoding="utf-8"
        ) as fd:
            for line in fd.readlines():
                line = line.strip()
                if line:
                    reqs.append(line)
        return reqs


def get_description():
    with codecs.open(os.path.join(BASE_DIR, "README.md"), "r", encoding="utf-8") as fh:
        return fh.read()


if __name__ == "__main__":
    setup(
        version=generate_version(),
        name="qtaf",
        packages=find_packages(
            exclude=(
                "tests",
                "tests.*",
            )
        ),
        py_modules=["qtaf_settings", "__main__", "qta-manage"],
        include_package_data=True,
        package_data={
            "": ["*.txt", "*.TXT"],
        },
        description="Basic test automation framework for QTA",
        long_description=get_description(),
        long_description_content_type="text/markdown",
        author="Tencent",
        license="Copyright(c)2010-2018 Tencent All Rights Reserved. ",
        install_requires=parse_requirements(),
        entry_points={
            "console_scripts": ["qta-manage = testbase.management:qta_manage_main"],
        },
        classifiers=[
            "Programming Language :: Python :: 2.7",
            "Operating System :: OS Independent",
        ],
        url="https://github.com/Tencent/QTAF",
        project_urls={
            "TestBase Documentation": "https://qta-testbase.readthedocs.io/zh/latest/",
            "TUIA Documentation": "https://qta-tuia.readthedocs.io/zh/latest/",
        },
    )



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""配置接口

一、配置格式
配置文件名默认为settings.py，可以通过环境变量指定用户配置文件的路径
qtaf相关配置的环境变量为：
QTAF_EXLIB_PATH: 指定exlib的路径，exlib下放置对应的qtaf、qt4s、installed_libs.txt等文件
QTAF_INSTALLED_LIBS: 指定已经安装的库的列表，以分号分隔，例如qt4s，用来代替不使用installed_libs.txt的场景
QTAF_SETTINGS_MODULE: 指定用户自定义的配置文件的路径，最后加载，会覆盖已经存在的配置

配置变量名必须符合pyton变量名规范，且统一使用大写，且不可以以"__"开头

如：
CONFIG_OPTION = True
DEBUG = True
RUNNER_THREAD_CNT = 5
BANNER = "hello"

二、使用示例：
from testbase.conf import settings
print(settings.CONFIG_OPTION)

注意 settings的值都是只读的，不可以修改，如果尝试修改会导致异常

三、配置优先级
配置存在2个优先级，当存在名字冲突时，使用高优先级的配置的值。优先级自低到高分别为：
1、QTAF配置                     固定为：test_proj/exlib/qtaf.egg/qtaf_settings.py
2、lib配置                      已经配置在test_proj/exlib/installed_libs.txt的包中的settings模块
3、用户自定义配置          固定为：test_proj/settings.py
"""

import imp
import os
import sys
import traceback

import qtaf_settings

from testbase import logger
from testbase.exlib import ExLibManager

_DEFAULT_SETTINSG_MODULE = "settings"


class _Settings(object):
    """配置读取接口"""

    def __init__(self):
        self.__keys = set()
        self.__sealed = False
        self.__loaded = False

    def _load(self):
        """加载配置
        :returns: Settings - 设置读取接口
        """
        # get PROJECT_MODE from user settings
        default_mode = getattr(qtaf_settings, "PROJECT_MODE", None)
        try:
            pre_settings = self._load_proj_settings_module("testbase.conf.pre_settings")
            mode = getattr(pre_settings, "PROJECT_MODE", default_mode)
        except ImportError as e:
            pre_settings = None
            if os.path.isfile(__file__):
                mode = "standard"
            else:
                mode = "standalone"

        if mode == "standard":  # library mode
            proj_root = getattr(pre_settings, "PROJECT_ROOT", os.getcwd())
            installed_apps = getattr(
                pre_settings,
                "INSTALLED_APPS",
                getattr(qtaf_settings, "INSTALLED_APPS", []),
            )

        else:  # egg mode
            proj_root = self._get_standalone_project_root(pre_settings)
            installed_apps = ExLibManager(proj_root).list_names()

        # load settings from qtaf
        self._load_setting_from_module(qtaf_settings)

        # load settings from installed apps
        for appname in installed_apps:
            modname = "%s.settings" % appname
            try:
                __import__(modname)
            except Exception:  # pylint: disable=broad-except
                stack = traceback.format_exc()
                logger.warn(
                    '[WARN]load library settings module "%s" failed:\n%s'
                    % (modname, stack)
                )
            else:
                self._load_setting_from_module(sys.modules[modname])

        # load settings from user settings
        try:
            proj_settings = self._load_proj_settings_module("testbase.conf.settings")
        except ImportError as e:
            if e.args[0] not in [
                "No module named settings",
                "No module named 'settings'",
            ]:
                # project settings found and there was an error
                stack = traceback.format_exc()
                logger.warn("[WARN]load project settings failed:\n%s" % stack)
        else:
            self._load_setting_from_module(proj_settings)

        # trying to set project root automatically
        setattr(self, "PROJECT_ROOT", proj_root)
        setattr(self, "INSTALLED_APPS", installed_apps)

    def _load_proj_settings_module(self, import_name):
        """加载项目配置文件"""
        user_settings = os.environ.get("QTAF_SETTINGS_MODULE", None)
        if user_settings:
            parts = user_settings.split(".")
            parts_temp = parts[:]
            dir_path = None
            while parts_temp:
                if dir_path:
                    fd, dir_path, desc = imp.find_module(parts_temp[0], [dir_path])
                else:
                    fd, dir_path, desc = imp.find_module(parts_temp[0])
                del parts_temp[0]
        else:
            fd, dir_path, desc = imp.find_module(_DEFAULT_SETTINSG_MODULE)
        return imp.load_module(import_name, fd, dir_path, desc)

    def _load_setting_from_module(self, module):
        """从模块中加载设置"""
        for name in dir(module):
            if name.startswith("__"):
                continue
            if name.islower():
                continue
            self.__keys.add(name)
            setattr(self, name, getattr(module, name))

    def _get_standalone_project_root(self, pre_settings):
        """获取独立模式下的项目的根目录"""
        proj_root = getattr(pre_settings, "PROJECT_ROOT", None)
        if proj_root:
            return proj_root
        if os.path.isfile(__file__):  # 没使用qtaf.egg包
            cwd = os.getcwd()
            # 使用外链或拷贝文件的方式
            dst_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
            if cwd.find(dst_path) >= 0:
                return os.path.abspath(dst_path)

            # eclipse调试使用工程引用的方式
            if "PYTHONPATH" not in os.environ:
                return cwd
            py_paths = os.environ["PYTHONPATH"]
            paths = py_paths.split(";")
            if len(paths) > 2:
                dst_path = paths[1]
            if cwd.find(dst_path) >= 0:
                return dst_path

            # 非预期的情况，返回当前工作目录
            return cwd
        else:  # 使用的egg包，qtaf.egg包在exlib目录中
            exlib_dir = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "..", ".."
            )
            proj_root = os.path.abspath(os.path.join(exlib_dir, ".."))
            return proj_root

    def get(self, name, *default_value):
        """获取配置"""
        if len(default_value) > 1:
            raise TypeError(
                "get expected at most 3 arguments, got %s" % (len(default_value) + 2)
            )
        if default_value:
            return getattr(self, name, default_value[0])
        else:
            return getattr(self, name)

    def __ensure_loaded(self):
        if not self.__loaded:
            self.__loaded = True
            self._load()
            self.__sealed = True

    def __setattr__(self, name, value):
        if not name.startswith("_Settings__") and self.__sealed:
            raise RuntimeError('尝试动态修改配置项"%s"' % name)
        else:
            super(_Settings, self).__setattr__(name, value)

    def __getattribute__(self, name):
        try:
            return super(_Settings, self).__getattribute__(name)
        except AttributeError:
            self.__ensure_loaded()
            return super(_Settings, self).__getattribute__(name)

    def __iter__(self):
        self.__ensure_loaded()
        return self.__keys.__iter__()

    def __contain__(self, key):
        self.__ensure_loaded()
        return key in self.__keys


settings = _Settings()


class _InnerSettings(object):
    """inner settings for a SettingsMixin class"""

    def __init__(self, defined_class):
        self.__tailer_names = {}
        self.__visited_settings_class = set()
        self.__allowed_prefix = set()
        self.__sealed = False
        self.__load_settings(defined_class)
        self.__sealed = True

    def __load_settings(self, defined_class):
        classes = [defined_class]
        base_classes = [defined_class.__bases__]
        visited_classes = set()
        while classes:
            temp_cls = classes[0]
            if not base_classes:
                if hasattr(temp_cls, "Settings"):
                    self.__allowed_prefix.add(temp_cls.__name__.upper())
                    self.__load_class_settings(temp_cls)
                visited_classes.add(temp_cls)
                classes.pop(0)
            else:
                if temp_cls in visited_classes:
                    base_classes = []
                else:
                    classes = list(temp_cls.__bases__) + classes
                    base_classes = list(classes[0].__bases__)

    def __load_class_settings(self, cls):
        prefix = cls.__name__.upper() + "_"
        class_path = cls.__module__ + "." + cls.__name__
        inner_settings_cls = getattr(cls, "Settings")
        if inner_settings_cls in self.__visited_settings_class:
            return

        self.__visited_settings_class.add(inner_settings_cls)
        for key in dir(inner_settings_cls):
            if key.startswith(prefix) and key.isupper():
                if key in settings:
                    value = settings.get(key)
                else:
                    value = getattr(inner_settings_cls, key)
                setattr(self, key, value)

                # handle legacy
                tailor_name = key[len(prefix) :]
                if tailor_name in self.__tailer_names:
                    for base_prefix in self.__tailer_names[tailor_name][:]:
                        base_name = base_prefix + tailor_name
                        setattr(self, base_name, value)
                        self.__tailer_names[tailor_name].append(prefix)
                else:
                    self.__tailer_names[tailor_name] = [prefix]

            elif not key.startswith("_"):
                for item in self.__allowed_prefix:
                    if key.startswith(item):
                        tailor_name = key[len(item) + 1 :]
                        break
                else:
                    tailor_name = key.upper()
                err_msg = "%s's Settings item `%s` must start with %s like %s%s" % (
                    class_path,
                    key,
                    prefix,
                    prefix,
                    tailor_name,
                )
                raise RuntimeError(err_msg)

    def __setattr__(self, name, value):
        if not name.startswith("_InnerSettings__") and self.__sealed:
            raise RuntimeError("dynamicly modifying value is not allowed.")
        super(_InnerSettings, self).__setattr__(name, value)

    def __getattribute__(self, name):
        try:
            return super(_InnerSettings, self).__getattribute__(name)
        except AttributeError:
            if name in settings:
                return settings.get(name)
            else:
                raise


class SettingsMixin(object):
    """a mixin class coordinate with qtaf settings"""

    @property
    def settings(self):
        cls = type(self)
        settings_key = "_%s_settings" % cls.__name__
        if not hasattr(cls, settings_key):
            if not hasattr(cls, "Settings"):
                raise RuntimeError("no inner class Settings defined")
            inner_settings = _InnerSettings(cls)
            setattr(cls, settings_key, inner_settings)
        return getattr(cls, settings_key)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""a module for unittest
"""
import os
import threading

from testbase.conf import settings


class _NotExistedItem(object):
    pass


class modify_settings(object):  # pylint: disable=invalid-name
    """temporarily modify settings"""

    def __init__(self, **kwargs):
        self.new_conf = kwargs
        self.old_conf = {}

    def __enter__(self):
        settings._Settings__ensure_loaded()  # ensure loaded
        settings._Settings__sealed = False
        for key in self.new_conf:
            self.old_conf[key] = getattr(settings, key, _NotExistedItem())
            if isinstance(self.old_conf[key], _NotExistedItem):
                settings._Settings__keys.add(key)
            setattr(settings, key, self.new_conf[key])

    def __exit__(self, *args):
        for key in self.old_conf:
            old_value = self.old_conf[key]
            if isinstance(old_value, _NotExistedItem):
                delattr(settings, key)
                settings._Settings__keys.remove(key)
            else:
                setattr(settings, key, old_value)
        settings._Settings__sealed = True


class modify_environ(object):  # pylint: disable=invalid-name
    """temporarily modify envrion"""

    def __init__(self, **kwargs):
        self.new_conf = kwargs
        self.old_conf = {}

    def __enter__(self):
        for key in self.new_conf:
            self.old_conf[key] = os.environ.get(key, _NotExistedItem())
            os.environ[key] = self.new_conf[key]

    def __exit__(self, *args):
        for key in self.old_conf:
            old_value = self.old_conf[key]
            if isinstance(old_value, _NotExistedItem):
                del os.environ[key]
            else:
                os.environ[key] = old_value


class modify_attributes(object):  # pylint: disable=invalid-name
    """temporarily modify attributes"""

    _lock = threading.Lock()

    def __init__(self, target, key_values):
        self._target = target
        self._old_values = {}
        self._new_values = key_values
        for key in key_values.keys():
            self._old_values[key] = getattr(target, key)

    def __enter__(self):
        self._lock.acquire()
        for key in self._new_values:
            setattr(self._target, key, self._new_values[key])

    def __exit__(self, *args):
        try:
            for key in self._old_values:
                setattr(self._target, key, self._old_values[key])
        finally:
            self._lock.release()



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
测试套基类模块
"""
from __future__ import absolute_import

import collections
import threading
import re
import traceback

import six

from testbase.testcase import ITestCaseRunner, TestCase, TestCaseRunner
from testbase.testresult import StreamResult, TestResultCollection
from testbase.util import ThreadGroupLocal, ThreadGroupScope


class SeqTestCaseRunner(ITestCaseRunner):
    """顺序执行的用例的执行器"""

    def run(self, testsuite, testresult_factory):
        """执行一个顺序执行的测试用例套

        :param testsuite: 执行的测试用例套
        :type testsuite: SeqTestSuite
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory

        :return TestResult/TestResultCollection - 测试结果
        """
        passed = True
        results = []
        for it in testsuite:
            runner = getattr(it, "case_runner", TestCaseRunner())
            case_result = runner.run(it, testresult_factory)
            passed &= case_result.passed
            results.append(case_result)
            if not passed:
                break
        return TestResultCollection(results, passed)


class TestSuiteBase(object):
    """测试套基类"""

    @property
    def suite_class_name(self):
        """测试套类名称"""
        cls = type(self)
        if cls.__module__ == "__main__":
            type_name = cls.__name__
        else:
            type_name = cls.__module__ + "." + cls.__name__
        return type_name

    def dumps(self):
        """序列化"""
        raise NotImplementedError()

    def loads(self, buf):
        """反序列化"""
        raise NotImplementedError()


class SeqTestSuite(TestSuiteBase):
    """顺序执行的测试用例套"""

    case_runner = SeqTestCaseRunner()

    def __init__(self, testcases):
        """构造函数

        :param testcases: 测试用例列表
        :type testcases: list
        :param name: 测试用例名
        :type name: string
        """
        self._testcases = testcases
        self._resmgr = None
        self.__share_data_mgr = None

    def __iter__(self):
        for it in self._testcases:
            yield it

    def __len__(self):
        return len(self._testcases)

    def __repr__(self):
        return "<SeqTestSuite module:%s>" % self.test_class_name

    @property
    def test_class_name(self):
        """返回测试用例名字（不同测试用例的名字不同）

        :rtype: str
        """
        cls = type(self._testcases[0])
        return cls.__module__

    @property
    def test_name(self):
        """返回测试用例实例的名字

        :rtype: str
        """
        cls = type(self._testcases[0])
        return cls.__module__

    @property
    def test_doc(self):
        """测试用例说明

        :rtype: str
        """
        cls = type(self._testcases[0])
        desc = cls.__module__.__doc__
        if isinstance(desc, six.text_type):
            desc = re.sub("^\s*", "", desc)
            desc = re.sub("\s*$", "", desc)
        return desc

    @property
    def test_result(self):
        """将最后一个执行的用例结果，作为Suite的结果"""
        result = None
        for testcae in self._testcases:
            if testcae.test_result:
                result = testcae.test_result
            else:
                break
        return result

    @property
    def test_resmgr(self):
        """资源管理器"""
        return self._resmgr

    @test_resmgr.setter
    def test_resmgr(self, resmgr):
        self._resmgr = resmgr
        for it in self._testcases:
            it.test_resmgr = resmgr

    @property
    def share_data_mgr(self):
        """共享数据管理器"""
        return self.__share_data_mgr

    @share_data_mgr.setter
    def share_data_mgr(self, share_data_mgr):
        self.__share_data_mgr = share_data_mgr
        for it in self._testcases:
            it._share_data_mgr = share_data_mgr

    def dumps(self):
        """序列化"""
        from testbase import serialization

        return [serialization.dumps(it) for it in self._testcases]

    def loads(self, buf):
        """反序列化"""
        from testbase import serialization

        self._testcases = [serialization.loads(it) for it in buf]


class TestSuiteCaseRunner(ITestCaseRunner):
    """测试用例套执行器（支持串/并行执行用例）"""

    def __init__(self, exec_mode, stop_on_failure=True, concurrency=1):
        self._exec_mode = exec_mode
        self._stop_on_failure = stop_on_failure
        self._concurrency = concurrency

    def _log_testsuite_error(self, testsuite, testresult, message):
        """记录测试用例套执行错误"""
        with ThreadGroupScope("%s:%s" % (testsuite.test_name, id(self))):

            ThreadGroupLocal().testcase = testsuite
            ThreadGroupLocal().testresult = testresult
            testresult.error(message)

    def _run_test(self, testsuite, test, testresult_factory, testresult):
        """执行测试用例"""
        testresult.begin_step(test.test_name)
        testsuite.current_stage = test.test_name
        runner = getattr(test, "case_runner", TestCaseRunner())
        case_result = runner.run(test, testresult_factory)
        if not case_result.passed:
            self._log_testsuite_error(
                testsuite, testresult, "TestCase %s run failed" % test.test_name
            )
        return case_result

    def sequential_run(
        self, testsuite, testresult_factory, testresult, stop_on_failure
    ):
        """顺序执行用例

        :param testsuite: 执行的测试用例套
        :type testsuite: SeqTestSuite
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :param stop_on_failure: 是否遇到失败用例停止执行
        :type stop_on_failure: bool

        :return TestResult/TestResultCollection - 测试结果
        """
        passed = True
        results = [testresult]
        for it in testsuite:
            case_result = self._run_test(testsuite, it, testresult_factory, testresult)
            passed &= case_result.passed
            results.append(case_result)
            if not passed and stop_on_failure:
                break
        return TestResultCollection(results, passed)

    def _run_test_from_queue(
        self,
        testsuite,
        tests_queue,
        testresult_factory,
        testresult,
        lock,
        test_result_dict,
    ):
        """从队列中不断取用例并执行

        :param tests_queue: 测试用例队列
        :type tests_queue: deque
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :param lock: 线程锁
        :type lock: threading.Lock
        :param test_result_dict: 测试结果字典
        :type test_result_dict: dict
        """
        while len(tests_queue) > 0:
            with lock:
                if len(tests_queue) <= 0:
                    break
                test = tests_queue.pop()
            case_result = self._run_test(
                testsuite, test, testresult_factory, testresult
            )
            with lock:
                test_result_dict.append(case_result)

    def parallel_run(self, testsuite, testresult_factory, testresult, concurrency):
        """并行执行用例

        :param testsuite: 执行的测试用例套
        :type testsuite: SeqTestSuite
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :param concurrency: 并发数
        :type concurrency: int

        :return TestResult/TestResultCollection - 测试结果
        """
        tests_queue = collections.deque([it for it in testsuite])
        results = [testresult]
        threads = []
        lock = threading.Lock()
        for _ in range(concurrency):
            thread = threading.Thread(
                target=self._run_test_from_queue,
                args=(
                    testsuite,
                    tests_queue,
                    testresult_factory,
                    testresult,
                    lock,
                    results,
                ),
            )
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()
        passed = True
        for result in results:
            if not result.passed:
                passed = False
                break
        return TestResultCollection(results, passed)

    def run(self, testsuite, testresult_factory):
        """执行一个顺序执行的测试用例套

        :param testsuite: 执行的测试用例套
        :type testsuite: SeqTestSuite
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory

        :return TestResult/TestResultCollection - 测试结果
        """
        if not isinstance(testsuite, TestSuite):
            raise ValueError("Invalid testsuite type: %s" % type(testsuite))
        testresult = testresult_factory.create(testsuite)
        testresult.begin_test(testsuite)
        testresult.begin_step("pre_test")
        testsuite.init_test(testresult)
        result = TestResultCollection([testresult], False)
        try:
            testsuite.pre_test()
        except:
            self._log_testsuite_error(
                testsuite,
                testresult,
                "pre_test run failed: \n%s" % traceback.format_exc(),
            )
        else:
            if self._exec_mode == TestSuite.EnumExecMode.Sequential:
                result = self.sequential_run(
                    testsuite,
                    testresult_factory,
                    testresult,
                    stop_on_failure=self._stop_on_failure,
                )
            elif self._exec_mode == TestSuite.EnumExecMode.Parallel:
                result = self.parallel_run(
                    testsuite,
                    testresult_factory,
                    testresult,
                    concurrency=self._concurrency,
                )
            else:
                raise ValueError("Invalid exec mode: %s" % self._exec_mode)

        testresult.begin_step("post_test")
        try:
            testsuite.post_test()
        except:
            self._log_testsuite_error(
                testsuite,
                testresult,
                "post_test run failed: \n%s" % traceback.format_exc(),
            )
        testresult.end_test()
        testsuite.end_test(result)
        return result


class TestSuite(TestSuiteBase):
    """测试套"""

    EnumPriority = TestCase.EnumPriority
    EnumStatus = TestCase.EnumStatus

    class EnumExecMode(object):
        Parallel = "parallel"
        Sequential = "sequential"

    owner = None
    priority = None
    status = None
    timeout = None
    testcases = []
    testcase_filter = {}
    exec_mode = EnumExecMode.Sequential
    stop_on_failure = True
    concurrency = 1

    def __init__(self, testcases):
        self.case_runner = TestSuiteCaseRunner(
            self.exec_mode,
            stop_on_failure=self.stop_on_failure,
            concurrency=self.concurrency,
        )
        self.__testcases = testcases
        self.__testresults = []
        self.__current_stage = ""

    def __iter__(self):
        for it in self.__testcases:
            yield it

    def __len__(self):
        return len(self.__testcases)

    @property
    def test_name(self):
        """测试套名字

        :return: str
        """
        return self.suite_class_name

    @property
    def test_doc(self):
        """测试套描述

        :return: str
        """
        return self.__class__.__doc__.strip()

    @property
    def test_result(self):
        """测试套结果

        :return: TestResult
        """
        if self.__testresults:
            return self.__testresults[0]
        return None

    @property
    def test_results(self):
        """测试用例结果

        :return: TestResultCollection
        """
        if len(self.__testresults) > 1:
            return self.__testresults[1:]
        return []

    @property
    def current_stage(self):
        return self.__current_stage

    @current_stage.setter
    def current_stage(self, value):
        self.__current_stage = value

    @classmethod
    def filter(cls, testcase):
        if not cls.testcase_filter:
            return False
        priority_list = cls.testcase_filter.get("priorities", [])
        if (
            priority_list
            and isinstance(priority_list, list)
            and testcase.priority not in priority_list
        ):
            return True
        status_list = cls.testcase_filter.get("statuses", [])
        if (
            status_list
            and isinstance(status_list, list)
            and testcase.status not in status_list
        ):
            return True
        owner_list = cls.testcase_filter.get("owners", [])
        if (
            owner_list
            and isinstance(owner_list, list)
            and testcase.owner not in owner_list
        ):
            return True
        tag_list = cls.testcase_filter.get("tags", [])
        if (
            tag_list
            and hasattr(testcase, "tags")
            and isinstance(tag_list, list)
            and set(tag_list).isdisjoint(testcase.tags)
        ):
            return True
        exclude_tag_list = cls.testcase_filter.get("exclude_tags", [])
        if (
            exclude_tag_list
            and isinstance(exclude_tag_list, list)
            and not set(exclude_tag_list).isdisjoint(testcase.tags)
        ):
            return True
        return False

    def add_share_data(self, name, value, level=0):
        """添加共享数据

        :type name: string
        :param name: 需要共享的数据名称
        :param value: 需要共享的数据内容
        """
        if self.share_data_mgr:
            self.share_data_mgr.set(name, value, level)

    def get_share_data(self, name):
        """从内存中获取存储的全局数据，给当前用例使用"""
        if self.share_data_mgr:
            return self.share_data_mgr.get(name)

    def remove_share_data(self, name):
        """从内存中移除存储的共享数据"""
        if self.share_data_mgr:
            return self.share_data_mgr.remove(name)

    def get_extra_fail_record(self):
        """当错误发生时，获取需要额外添加的日志记录和附件信息

        :return: dict,dict - 日志记录，附件信息
        """
        return {}, {}

    def get_test_extra_properties(self):
        return {}

    def init_test(self, testresult):
        """初始化测试套。慎用此函数，尽量将初始化放到pre_test里。

        :param testresult: 测试结果
        :type testresult: TestResult
        """
        self.__testresults.append(testresult)

    def pre_test(self):
        pass

    def post_test(self):
        pass

    def end_test(self, testresults):
        self.__testresults = testresults

    def dumps(self):
        """序列化"""
        from testbase import serialization

        return [serialization.dumps(it) for it in self.__testcases]

    def loads(self, buf):
        """反序列化"""
        from testbase import serialization

        self.__testcases = [serialization.loads(it) for it in buf]

    def debug_run(self):
        """本地调试测试套"""
        from testbase.runner import TestRunner
        from testbase.report import EmptyTestReport

        tests = [self]
        report = EmptyTestReport(lambda tc: StreamResult())
        runner = TestRunner(report)
        return runner.run(tests)
